from .GLConfig import GLConfig

class TextureUnits:
    def __init__(self):
        self._unit_texture_map = {}
        self._texture_unit_map = {}

    def __getitem__(self, unit:int)->tuple:
        key = (GLConfig.buffered_current_context, unit)
        if key not in self._unit_texture_map:
            return None
        
        return self._unit_texture_map[key]
    
    def __setitem__(self, unit:int, texture:tuple):
        context = GLConfig.buffered_current_context

        unit_key = (context, unit)
        if unit_key in self._unit_texture_map:
            old_texture = self._unit_texture_map[unit_key]
            if old_texture != texture:
                del self._texture_unit_map[context, old_texture]

        texture_key = (context, texture)
        if texture_key in self._texture_unit_map:
            old_unit = self._texture_unit_map[texture_key]
            if old_unit != unit:
                del self._unit_texture_map[context, old_unit]

        self._unit_texture_map[unit_key] = texture
        self._texture_unit_map[texture_key] = unit

    @property
    def current_texture(self):
        return self[GLConfig.active_texture_unit]
    
    @current_texture.setter
    def current_texture(self, texture:tuple):
        self[GLConfig.active_texture_unit] = texture

    def unit_of_texture(self, texture:tuple)->int:
        key = (GLConfig.buffered_current_context, texture)
        if key not in self._texture_unit_map:
            return None
        
        return self._texture_unit_map[key]
    
    @property
    def available_unit(self):
        context = GLConfig.buffered_current_context

        for i in range(GLConfig.max_texture_units):
            key = (context, i)
            if key not in self._unit_texture_map:
                return i
            
            if self._unit_texture_map[key] == 0:
                return i
            
        return None

TextureUnits = TextureUnits()



import cv2
import numpy as np
import glm
from OpenGL import GL
import OpenGL.GL.ARB.bindless_texture as bt

from .FBOAttachment import FBOAttachment
from .GLInfo import GLInfo
from .utils import checktype
from .helper import get_dtype, get_channels, width_adapt, get_external_format
from .ImageLoader import ImageLoader
from .DictList import DictList

class samplerCube(FBOAttachment):

	_basic_info = \
	{
		"gen_func": GL.glGenTextures,
		"bind_func": GL.glBindTexture,
		"del_func": GL.glDeleteTextures,
		"target_type": GL.GL_TEXTURE_CUBE_MAP,
		"binding_type": GL.GL_TEXTURE_BINDING_CUBE_MAP,
		"need_number": True,
	}

	class Face:
		def __init__(self, sampler_cube, texture_target, width:int=0, height:int=0, internal_format:GLInfo.internal_formats=GL.GL_RGBA32F):
			self._texture_target = texture_target
			self._sampler_cube = sampler_cube
			self._dynamic = sampler_cube.dynamic

			self._image = None
			self._width = width
			self._height = height
			self._internal_format = internal_format
			self._dtype = None

			self._image_changed = True

		@property
		def dynamic(self):
			return self._dynamic

		@property
		def width(self):
			return self._width

		@width.setter
		@FBOAttachment.param_setter
		def width(self, width:int):
			self._width = width
			if self._image is not None:
				if len(self._image.shape) > 2:
					cv2.resize(self._image, (self._image.shape[0], width, self._image.shape[2]))
				else:
					cv2.resize(self._image, (self._image.shape[0], width))

			self._image_changed = True

		@property
		def height(self):
			return self._height

		@height.setter
		@FBOAttachment.param_setter
		def height(self, height:int):
			self._height = height
			if self._image is not None:
				if len(self._image.shape) > 2:
					cv2.resize(self._image, (height, self._image.shape[1], self._image.shape[2]))
				else:
					cv2.resize(self._image, (height, self._image.shape[1]))

			self._image_changed = True

		@property
		def dtype(self):
			return get_dtype(self._internal_format)

		@property
		def internal_format(self):
			return self._internal_format

		@internal_format.setter
		@FBOAttachment.param_setter
		def internal_format(self, format:GLInfo.internal_formats):
			if self._internal_format == format:
				return
			
			self._internal_format = format
			if self._image is not None:
				old_channels = self._image.shape[2] if len(self._image.shape) > 2 else 1
				new_channels = get_channels(format)
				if old_channels == 3 and new_channels == 4:
					self._image = cv2.cvtColor(self._image, cv2.COLOR_RGB2RGBA)
				elif old_channels == 4 and new_channels == 3:
					self._image = cv2.cvtColor(self._image, cv2.COLOR_RGBA2RGB)
				elif old_channels != new_channels:
					self._image = None
			if self._image is not None:
				old_np_dtype = self._image.dtype
				new_np_dtype = GLInfo.dtype_map[get_dtype(self._internal_format)]
				if old_np_dtype != new_np_dtype:
					self._image = self._image.astype(new_np_dtype)
			self._image_changed = True

		@property
		def image(self):
			return self._image

		@image.setter
		@FBOAttachment.param_setter
		def image(self, image:(np.ndarray,str)):
			if isinstance(image, str):
				image = ImageLoader.load(image)

			self._image = image
			
			self._width = self._image.shape[1]
			self._height = self._image.shape[0]
			channels = self._image.shape[2] if len(self._image.shape) > 2 else 1
			self._internal_format = GLInfo.internal_formats_map[self._image.dtype][channels]
			self._dtype = GLInfo.dtype_inverse_map[self._image.dtype]

			self._image_changed = True

		@property
		def target(self):
			return self._texture_target

		def _apply(self, force_update_image:bool=False):
			if not self._image_changed and not force_update_image:
				return False
			
			width_adapt(self._width)
			external_format = get_external_format(self._internal_format)
			GL.glTexImage2D(self._texture_target, 0, self._internal_format,
							self._width, self._height, 0, external_format, self.dtype, self._image)
			self._image_changed = False

			return True

		@checktype
		def malloc(self, width:int, height:int, internal_format:GLInfo.internal_formats=None):
			self.width = width
			self.height = height

			if internal_format is None:
				internal_format = self.internal_format

			if internal_format is None:
				internal_format = GL.GL_RGBA32F

			self.internal_format = internal_format

	def __init__(self, width:int=0, height:int=0, internal_format:GLInfo.internal_formats=GL.GL_RGBA32F):
		FBOAttachment.__init__(self)

		self._handle = 0
		self._dynamic = True

		self._faces = DictList()
		self._faces["right"] = samplerCube.Face(self, GL.GL_TEXTURE_CUBE_MAP_POSITIVE_X, width, height, internal_format)
		self._faces["left"] = samplerCube.Face(self, GL.GL_TEXTURE_CUBE_MAP_NEGATIVE_X, width, height, internal_format)
		self._faces["bottom"] = samplerCube.Face(self, GL.GL_TEXTURE_CUBE_MAP_POSITIVE_Y, width, height, internal_format)
		self._faces["top"] = samplerCube.Face(self, GL.GL_TEXTURE_CUBE_MAP_NEGATIVE_Y, width, height, internal_format)
		self._faces["front"] = samplerCube.Face(self, GL.GL_TEXTURE_CUBE_MAP_POSITIVE_Z, width, height, internal_format)
		self._faces["back"] = samplerCube.Face(self, GL.GL_TEXTURE_CUBE_MAP_NEGATIVE_Z, width, height, internal_format)

		self._wrap_s = GL.GL_CLAMP_TO_EDGE
		self._wrap_t = GL.GL_CLAMP_TO_EDGE
		self._wrap_r = GL.GL_CLAMP_TO_EDGE
		self._filter_min = GL.GL_LINEAR
		self._filter_mag = GL.GL_LINEAR
		self._filter_mipmap = GL.GL_LINEAR
		self._border_color = glm.vec4(0, 0, 0, 1)

		self._wrap_s_changed = True
		self._wrap_t_changed = True
		self._wrap_r_changed = True
		self._filter_min_changed = True
		self._filter_mag_changed = True
		self._border_color_changed = False

	def __del__(self):
		if self._handle != 0:
			self._handle = 0

		FBOAttachment.__del__(self)

	def __getitem__(self, key:(str,int)):
		if key == "up":
			key = "top"
		elif key == "down":
			key = "bottom"

		return self._faces[key]

	@checktype
	def __setitem__(self, key:str, value:(np.ndarray,str)):
		if key == "up":
			key = "top"
		elif key == "down":
			key = "bottom"

		self._faces[key].image = value

	def __getattr__(self, key):
		if key not in ["right", "left", "top", "bottom", "back", "front", "up", "down"]:
			return super().__getattribute__(key)

		return self[key]

	def __setattr__(self, key, value):
		if key not in ["right", "left", "top", "bottom", "back", "front", "up", "down"]:
			return super().__setattr__(key, value)

		self[key] = value

	@property
	def handle(self):
		if not bt.glGetTextureHandleARB:
			return 0

		if self._id == 0:
			return 0
		
		if self._handle == 0:
			self._handle = bt.glGetTextureHandleARB(self._id)
			if self._handle == 0:
				raise RuntimeError("failed to create samplerCube {self._id}'s handle")
			bt.glMakeTextureHandleResidentARB(self._handle)
			self._dynamic = False

		return self._handle

	@property
	def is_completed(self):
		for face in self._faces:
			if face.width <= 0 or face.height <= 0:
				return False
			
		return True

	def bind(self, update_fbo:bool=False, force_update_image:bool=False):
		FBOAttachment.bind(self, update_fbo, force_update_image)
		GL.glEnable(GL.GL_TEXTURE_CUBE_MAP_SEAMLESS)

		if self._wrap_s_changed:
			GL.glTexParameteri(GL.GL_TEXTURE_CUBE_MAP, GL.GL_TEXTURE_WRAP_S, self._wrap_s)
			self._wrap_s_changed = False
		if self._wrap_t_changed:
			GL.glTexParameteri(GL.GL_TEXTURE_CUBE_MAP, GL.GL_TEXTURE_WRAP_T, self._wrap_t)
			self._wrap_t_changed = False
		if self._wrap_r_changed:
			GL.glTexParameteri(GL.GL_TEXTURE_CUBE_MAP, GL.GL_TEXTURE_WRAP_R, self._wrap_r)
			self._wrap_r_changed = False
		if self._border_color_changed:
			GL.glTexParameterfv(GL.GL_TEXTURE_CUBE_MAP, GL.GL_TEXTURE_BORDER_COLOR, glm.value_ptr(self._border_color))
			self._border_color_changed = False

		if self._filter_min_changed:
			GL.glTexParameteri(GL.GL_TEXTURE_CUBE_MAP, GL.GL_TEXTURE_MIN_FILTER, self._get_filter_min())
			self._filter_min_changed = False
		if self._filter_mag_changed:
			GL.glTexParameteri(GL.GL_TEXTURE_CUBE_MAP, GL.GL_TEXTURE_MAG_FILTER, self._filter_mag)
			self._filter_mag_changed = False

		image_changed = False
		for face in self._faces:
			image_changed = (face._apply(force_update_image) or image_changed)

		if not update_fbo and self._filter_mipmap is not None:
			generated_mipmap = False
			if image_changed:
				self.generate_mipmap()
				generated_mipmap = True

			if not generated_mipmap and not self._fbo_image_generated_mipmap:
				self.generate_mipmap()

	def generate_mipmap(self):
		FBOAttachment.bind(self)
		GL.glGenerateMipmap(GL.GL_TEXTURE_CUBE_MAP)
		self._fbo_image_generated_mipmap = True

	@property
	def internal_format(self):
		return self._faces[0].internal_format
	
	@internal_format.setter
	@FBOAttachment.param_setter
	def internal_format(self, internal_format:GLInfo.internal_formats):
		for i in range(6):
			self._faces[i].internal_format = internal_format

	@property
	def wrap_s(self):
		return self._wrap_s

	@wrap_s.setter
	@FBOAttachment.param_setter
	def wrap_s(self, wrap_type:GLInfo.wrap_types):
		self._wrap_s = wrap_type
		self._wrap_s_changed = True

	@property
	def wrap_t(self):
		return self._wrap_t

	@wrap_t.setter
	@FBOAttachment.param_setter
	def wrap_t(self, wrap_type:GLInfo.wrap_types):
		self._wrap_t = wrap_type
		self._wrap_t_changed = True

	@property
	def wrap_r(self):
		return self._wrap_r

	@wrap_r.setter
	@FBOAttachment.param_setter
	def wrap_r(self, wrap_type:GLInfo.wrap_types):
		self._wrap_r = wrap_type
		self._wrap_r_changed = True

	@property
	def wrap(self):
		return self._wrap_s, self._wrap_t, self._wrap_r

	@wrap.setter
	@FBOAttachment.param_setter
	def wrap(self, wrap_type):
		if wrap_type in GLInfo.wrap_types:
			self.wrap_s = wrap_type
			self.wrap_t = wrap_type
			self.wrap_r = wrap_type
		else:
			self.wrap_s = wrap_type[0]
			self.wrap_t = wrap_type[1]
			self.wrap_r = wrap_type[2]

	@property
	def filter_min(self):
		return self._filter_min

	@filter_min.setter
	@FBOAttachment.param_setter
	def filter_min(self, filter_type:GLInfo.filter_types):
		self._filter_min = filter_type
		self._filter_min_changed = True

	@property
	def filter_mag(self):
		return self._filter_mag

	@filter_mag.setter
	@FBOAttachment.param_setter
	def filter_mag(self, filter_type:GLInfo.filter_types):
		self._filter_mag = filter_type
		self._filter_mag_changed = True

	@property
	def filter_mipmap(self):
		return self._filter_mipmap

	@filter_mipmap.setter
	@FBOAttachment.param_setter
	def filter_mipmap(self, filter_type:GLInfo.filter_types):
		self._filter_mipmap = filter_type
		self._filter_min_changed = True
		self._filter_mag_changed = True

	def _get_filter_min(self):
		if self._filter_mipmap is None:
			return self._filter_min
		else:
			return GLInfo.mipmap_filter_map[(self._filter_min, self._filter_mipmap)]
		
	@property
	def filter(self):
		return self._filter_min, self._filter_mag, self._filter_mipmap

	@filter.setter
	@FBOAttachment.param_setter
	def filter(self, filter_type):
		if filter_type in GLInfo.filter_types:
			self.filter_min = filter_type
			self.filter_mag = filter_type
		else:
			self.filter_min = filter_type[0]
			self.filter_mag = filter_type[1]

	@property
	def border_color(self):
		return self._border_color

	@border_color.setter
	@FBOAttachment.param_setter
	def border_color(self, color:(glm.vec3, glm.vec4)):
		if isinstance(color, glm.vec3):
			color = glm.vec4(color, 1)
		
		self._border_color = color
		self._border_color_changed = True

	@checktype
	def malloc(self, width:int, height:int, samples:int=None, layers:int=None, internal_format:GLInfo.internal_formats=None):
		for face in self._faces:
			face.malloc(width, height, internal_format)

	@property
	def width(self):
		return self.faces[0].width
	
	@property
	def height(self):
		return self.faces[0].height
	
	@property
	def layers(self):
		return 6

	@property
	def faces(self):
		return self._faces

	@property
	def samples(self):
		return None
	
	@samples.setter
	def samples(self, samples:int):
		pass


import glm
from glass import Instance

class AffineTransform(Instance):
	def __init__(self):
		Instance.__init__(self)
		
		self["affine_transform_row0"] = glm.vec4(1, 0, 0, 0)
		self["affine_transform_row1"] = glm.vec4(0, 1, 0, 0)
		self["affine_transform_row2"] = glm.vec4(0, 0, 1, 0)
		self["env_map_handle"] = 0
		self["visible"] = 1

	def apply(self, pos):
		transform_mat = glm.mat4(self["affine_transform_row0"],
			                      self["affine_transform_row1"],
								  self["affine_transform_row2"],
								  glm.vec4(0, 0, 0, 1))
		transform_mat = glm.transpose(transform_mat)
		result = transform_mat * glm.vec4(*pos, 1)
		return result.xyz


from ..Mesh import Mesh

from glass.utils import checktype
from glass import Vertex
from glass.ImageLoader import ImageLoader

import glm
import numpy as np
from enum import Enum

class ImageQuad(Mesh):

    class AlignCenter(Enum):
        Center = (0, 0)
        LeftCenter = (-0.5, 0)
        RightCenter = (0.5, 0)
        TopCenter = (0, 0.5)
        BottomCenter = (0, -0.5)
        LeftTop = (-0.5, 0.5)
        RightTop = (0.5, 0.5)
        LeftBottom = (-0.5, -0.5)
        RightBottom = (0.5, -0.5)

    @checktype
    def __init__(self, image:(str,np.ndarray), align_center:AlignCenter=AlignCenter.BottomCenter,
                 width:float=None, height:float=None,
                 vertical:bool=True, name:str=""):
        Mesh.__init__(self, color=glm.vec4(0), back_color=glm.vec4(0), name=name, block=True)
        self.__align_center = align_center

        _image = image
        aspect = 1
        if isinstance(image, str):
            try:
                _image = ImageLoader.load(image)
                aspect = _image.shape[1] / _image.shape[0]
            except:
                pass
        elif isinstance(image, np.ndarray):
            aspect = _image.shape[1] / _image.shape[0]

        if width is None and height is None:
            width = 1

        if width is None:
            width = height * aspect

        if height is None:
            height = width / aspect

        self.__width = width
        self.__height = height
        self.__vertical = vertical
        
        self.material.diffuse_map = image
        self.start_building()

    def build(self):
        self.is_closed = False
        self.self_calculated_normal = True

        vertices = self.vertices
        indices = self.indices        
        width = self.__width
        height = self.__height
        vertical = self.__vertical
        align0 = self.__align_center.value[0]*width
        align1 = self.__align_center.value[1]*height

        normal = glm.vec3(0, 0, 1)
        if vertical:
            normal = glm.vec3(0, -1, 0)

        # 左下
        vertex0 = Vertex()
        vertex0.position = glm.vec3(-width/2-align0, -height/2-align1, 0)
        if vertical:
            vertex0.position = glm.vec3(-width/2-align0, 0, -height/2-align1)
        vertex0.normal = normal
        vertex0.tex_coord = glm.vec3(0)

        # 右下
        vertex1 = Vertex()
        vertex1.position = glm.vec3(width/2-align0, -height/2-align1, 0)
        if vertical:
            vertex1.position = glm.vec3(width/2-align0, 0, -height/2-align1)
        vertex1.normal = normal
        vertex1.tex_coord = glm.vec3(1, 0, 0)

        # 右上
        vertex2 = Vertex()
        vertex2.position = glm.vec3(width/2-align0, height/2-align1, 0)
        if vertical:
            vertex2.position = glm.vec3(width/2-align0, 0, height/2-align1)
        vertex2.normal = normal
        vertex2.tex_coord = glm.vec3(1, 1, 0)

        # 左上
        vertex3 = Vertex()
        vertex3.position = glm.vec3(-width/2-align0, height/2-align1, 0)
        if vertical:
            vertex3.position = glm.vec3(-width/2-align0, 0, height/2-align1)
        vertex3.normal = normal
        vertex3.tex_coord = glm.vec3(0, 1, 0)

        vertices[0] = vertex0
        vertices[1] = vertex1
        vertices[2] = vertex2
        vertices[3] = vertex3

        indices[0] = glm.uvec3(0, 1, 2)
        indices[1] = glm.uvec3(0, 2, 3)

        self.generate_temp_TBN(vertices[0], vertices[1], vertices[2])
        self.generate_temp_TBN(vertices[0], vertices[2], vertices[3])

        del vertices[4:]
        del indices[2:]

    @property
    def vertical(self):
        return self.__vertical
    
    @vertical.setter
    @Mesh.param_setter
    def vertical(self, flag:bool):
        self.__vertical = flag

    @property
    def align_center(self):
        return self.__align_center
    
    @align_center.setter
    @Mesh.param_setter
    def align_center(self, align_center:AlignCenter):
        self.__align_center = align_center

    @property
    def width(self):
        return self.__width
    
    @width.setter
    @Mesh.param_setter
    def width(self, width:float):
        self.__width = width

    @property
    def height(self):
        return self.__height
    
    @height.setter
    @Mesh.param_setter
    def height(self, height:float):
        self.__height = height



from .PostProcessEffects import PostProcessEffects
from .PostProcessEffect import PostProcessEffect
from .ShaderEffect import ShaderEffect

from .GaussBlur import GaussBlur
from .KernelFilter import KernelFilter
from .BloomEffect import BloomEffect
from .ACESToneMapper import ACESToneMapper
from .DOFEffect import DOFEffect
from .FXAAEffect import FXAAEffect
from .LUTEffect import LUTEffect
from .SSAOEffect import SSAOEffect
from .MulFilter import MulFilter
from .ExplosureAdaptor import ExplosureAdaptor


from .utils import di

class WeakRef:

    def __init__(self, obj):
        if isinstance(obj, (int,float,bool,str,bytes,type(None))):
            self._obj_id = obj
            self._is_real_id = False
        else:
            self._obj_id = id(obj)
            self._is_real_id = True

    def __hash__(self):
        return self._obj_id
    
    def __eq__(self, other):
        return (self._obj_id == other._obj_id)
    
    def __repr__(self):
        return self.value.__repr__()

    @property
    def value(self):
        if self._is_real_id:
            return di(self._obj_id)
        else:
            return self._obj_id
        


from functools import wraps
import ctypes
import re
import math
import glm
import os
import types
import cProfile
import hashlib
import pickle
import subprocess
import sys
import wget
import chardet
import requests
import moderngl
from geolite2 import geolite2
from _ctypes import PyObj_FromPtr

from .GlassConfig import GlassConfig

profiler = cProfile.Profile()

def run_exe(exe_path, *args):
    cmd = [exe_path, *args]
    result = subprocess.run(cmd, capture_output=True, text=True)
    stdout = result.stdout.strip()
    stderr = result.stderr.strip()
    return stdout, stderr

def cat(file_name):
    in_file = open(file_name, "r", encoding=sys.getdefaultencoding(), errors="ignore")
    content = in_file.read()
    in_file.close()
    return content

def bincat(file_name):
    in_file = open(file_name, "rb")
    content = in_file.read()
    in_file.close()
    return content

def uint64_to_uvec2(uint64_value):
    result = glm.uvec2()
    bytes_data = uint64_value.to_bytes(8, byteorder='little', signed=False)
    result.x = int.from_bytes(bytes_data[:4], byteorder='little', signed=False)
    result.y = int.from_bytes(bytes_data[4:], byteorder='little', signed=False)
    return result

def same_type(var1, var2):
    return (isinstance(var1, type(var2)) and isinstance(var2, type(var1)))

def checktype(func):
    if not GlassConfig.debug:
        return func
    
    @wraps(func)
    def wrapper(*args, **kwargs):
        if not GlassConfig.debug:
            return func(*args, **kwargs)

        func_args = func.__code__.co_varnames[:func.__code__.co_argcount]
        annotation = func.__annotations__
        defaults = func.__defaults__
        len_func_args = len(func_args)

        len_defaults = 0
        if defaults is not None:
            len_defaults = len(defaults)
        defauts_start = len_func_args - len_defaults

        ext_kwargs = {}
        for i in range(defauts_start, len_func_args):
            ext_kwargs[func_args[i]] = defaults[i-defauts_start]
        ext_kwargs.update(kwargs)

        def change_annotation(arg_name):
            if annotation[arg_name] is None:
                annotation[arg_name] = (type(None),)

            if isinstance(annotation[arg_name], type):
                annotation[arg_name] = (annotation[arg_name], type(None))

            if float in annotation[arg_name] and \
                int not in annotation[arg_name]:
                annotation[arg_name] = (*annotation[arg_name], int)

            if None in annotation[arg_name]:
                type_set = set(annotation[arg_name])
                type_set.remove(None)
                type_set.add(type(None))
                annotation[arg_name] = tuple(type_set)

            if type(None) not in annotation[arg_name]:
                annotation[arg_name] = (*annotation[arg_name], type(None))

        for i in range(min(len(args), len_func_args)):
            arg_name = func_args[i]
            arg_value = args[i]

            if arg_name in annotation:
                if isinstance(annotation[arg_name], (type,tuple,type(None))):
                    change_annotation(arg_name)
                    if not isinstance(arg_value, annotation[arg_name]):
                        raise TypeError(func.__name__ + "() argument " + arg_name + " should be in type " + str(annotation[arg_name]) + \
                                        ", " + str(type(arg_value)) + " value passed.")
                elif isinstance(annotation[arg_name], (list,set)):
                    if arg_value not in annotation[arg_name]:
                        raise ValueError(func.__name__ + "() argument " + arg_name + " can only be in " + str(annotation[arg_name]) + \
                                         ", " + str(arg_value) + " passed.")

        for arg_name in ext_kwargs:
            arg_value = ext_kwargs[arg_name]
            if arg_name in annotation:
                if isinstance(annotation[arg_name], (type,tuple,type(None))):
                    change_annotation(arg_name)
                    if not isinstance(arg_value, annotation[arg_name]):
                        raise TypeError(func.__name__ + "() argument " + arg_name + " should be in type " + str(annotation[arg_name]) + \
                                        ", " + str(type(arg_value)) + " value passed.")
                elif isinstance(annotation[arg_name], (list,set)):
                    if arg_value not in annotation[arg_name]:
                        raise ValueError(func.__name__ + "() argument " + arg_name + " can only be in " + str(annotation[arg_name]) + \
                                         ", " + str(arg_value) + " passed.")

        return_value = func(*args, **kwargs)
        if "return" in annotation:
            if isinstance(annotation["return"], (type,tuple,type(None))):
                change_annotation("return")
                if not isinstance(return_value, annotation["return"]):
                    raise TypeError(func.__name__ + "() return value should be in type " + str(annotation["return"]) + \
                                    ", " + str(type(return_value)) + " value returned.")
            elif isinstance(annotation["return"], (list,set)):
                if return_value not in annotation["return"]:
                    raise ValueError(func.__name__ + "() return value can only be in " + str(annotation["return"]) + \
                                     ", " + str(return_value) + " returned.")

        return return_value

    return wrapper

def public_ip():
    try:
        response = requests.get('https://httpbin.org/ip')
        return response.json().get('origin')
    except:
        return None

def is_China_ip(ip_address):
    reader = geolite2.reader()
    location = reader.get(ip_address)
    country = location.get('country', {}).get('iso_code') if location else None
    return country == 'CN'

def pip_install(package_name:str):
    install_cmd = [sys.executable, "-m", "pip", "install", package_name]
    if "/" not in package_name and is_China_ip(public_ip()):
        install_cmd = [sys.executable, "-m", "pip", "install", package_name, "-i", "https://pypi.tuna.tsinghua.edu.cn/simple"]
    
    return_code = subprocess.call(install_cmd)
    if return_code != 0:
        if "/" not in package_name:
            raise RuntimeError(f"failed to install {package_name}")

        target_file = GlassConfig.cache_folder + "/" + os.path.basename(package_name)
        wget.download(package_name, target_file)
        install_cmd = [sys.executable, "-m", "pip", "install", target_file]
        return_code = subprocess.call(install_cmd)
        if return_code != 0:
            raise RuntimeError(f"failed to install {package_name}")

def is_overridden(method):
    cls = method.__self__.__class__
    mro = cls.__mro__
    code = method.__code__
    method_name = method.__name__
    for i, base_cls in enumerate(mro):
        if method_name in base_cls.__dict__:
            base_method = base_cls.__dict__[method_name]
            if isinstance(base_method, types.FunctionType) and base_method.__code__ is not code:
                return True
    return False

def delete(method):
    @wraps(method)
    def wrapper(*args, **kwargs):
        self = args[0]
        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{method.__name__}'")
    return wrapper

def extname(filename):
    return os.path.splitext(filename)[1][1:].lower()

def has_valid(content):
    its = list(re.finditer(r"\S", content))
    return its.__bool__()

def rget_token(content, i):
    token = {}

    if i < 0 or i >= len(content):
        return "", -1

    i = rskip_space(content, i)
    if i < 0:
        return "", -1

    pos_start = i
    i = rskip_valid(content, i)
    pos_end = i

    token["span"] = (pos_end+1, pos_start+1)
    token["word"] = content[pos_end+1:pos_start+1]

    return token, i

def reg_rfind(content, pattern, pos_start, pos_end):
    str_index, comment_index = get_invalid_index(content)

    its = list(re.finditer(pattern, content[pos_start:pos_end]))
    if not its:
        return -1

    for it in reversed(its):
        if it.start() not in str_index and it.start() not in comment_index:
            return it.start()

    return it.start()

def rskip_space(content, i):
    if i < 0 or i >= len(content):
        return -1

    while True:
        if i < 0 or content[i] not in " \t\r.":
            break
        i -= 1

    return i

def rskip_valid(content, i):
    if i < 0 or i >= len(content):
        return -1

    if content[i] in "+-*/%!=<>,":
        i -= 1
        return i

    str_index, comment_index = get_invalid_index(content)
    while i >= 0:
        if i not in str_index and i not in comment_index:
            if content[i] in "\n\t+-*/%!=<>,([{":
                break

            if content[i] in " \t":
                j = rskip_space(content, i)
                if j == -1:
                    break

                if content[j] not in ")]}":
                    break
                else:
                    i = rfind_pair(content, j)

            elif content[i] in ")]}":
                i = rfind_pair(content, i)

        i -= 1

    return i

def find_pair(content, i):
    len_content = len(content)
    if i < 0 or i >= len_content or content[i] not in "([{":
        return -1

    start_pair = content[i]
    end_pair = None
    if content[i] == '(':
        end_pair = ')'
    elif content[i] == '[':
        end_pair = ']'
    elif content[i] == '{':
        end_pair = '}'

    n_start_pair = 0
    str_index, comment_index = get_invalid_index(content)
    while i < len_content:
        if i not in str_index and i not in comment_index:
            if content[i] == start_pair:
                n_start_pair += 1
            elif content[i] == end_pair:
                n_start_pair -= 1

            if n_start_pair == 0:
                break

        i += 1

    return i

def is_text_file(file_path):
    try:
        with open(file_path, 'rb') as f:
            result = chardet.detect(f.read())
            return (result['encoding'] is not None)
    except:
        return False

def is_url(line:str)->bool:
    pattern = re.compile(r'http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\\(\\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+')
    return bool(pattern.match(line))

def md5s(content):
    md5_hash = hashlib.md5()
    if isinstance(content, str):
        md5_hash.update(content.encode('utf-8'))
    elif isinstance(content, (bytes,bytearray)):
        md5_hash.update(content)
    else:
        md5_hash.update(pickle.dumps(content))

    return md5_hash.hexdigest()

def modify_time(file_name):
    if not os.path.isfile(file_name):
        return 0
    
    return os.path.getmtime(file_name)

def relative_path(file_name, start_path="."):
    try:
        return os.path.relpath(file_name, start_path).replace("\\", "/")
    except:
        return os.path.abspath(file_name).replace("\\", "/")

def save_var(var, file_path):
    with open(file_path, 'wb') as file:
        pickle.dump(var, file)

def load_var(file_path):
    with open(file_path, 'rb') as file:
        var = pickle.load(file)
    return var

def subscript(var, subscript_chain, feed_index:int=None):
    for operator, operant in subscript_chain:
        if operator == "getattr":
            var = getattr(var, operant)
        else:
            used_index = (operant if operant != "{0}" else feed_index)
            var = var[used_index]

    return var

def subscript_set(var, subscript_chain, value, feed_index:int=None, compare_before_set:bool=True):
    len_chain = len(subscript_chain)
    last_index = len_chain - 1
    for i in range(len_chain):
        item = subscript_chain[i]
        operator = item[0]
        operant = item[1]
        
        if operator == "getattr":
            old_value = getattr(var, operant)
            if i != last_index:
                var = old_value
            elif not compare_before_set or old_value != value:
                setattr(var, operant, value)
        else:
            used_index = (operant if operant != "{0}" else feed_index)
            old_value = var[used_index]
            if i != last_index:
                var = old_value
            elif not compare_before_set or old_value != value:
                var[used_index] = value

def get_subscript_chain(suffix):
    subscript_chain = []
    len_suffix = len(suffix)
    pos_start = 0
    while pos_start < len_suffix:
        start_char = suffix[pos_start]
        if start_char == "." or \
        'a' <= start_char <= 'z' or \
        'A' <= start_char <= 'Z':
            if start_char == ".":
                pos_start += 1

            pos_point = suffix.find(".", pos_start)
            if pos_point == -1:
                pos_point = len_suffix

            pos_braket = suffix.find("[", pos_start)
            if pos_braket == -1:
                pos_braket = len_suffix

            pos_end = min(pos_point, pos_braket)
            attr_name = suffix[pos_start:pos_end]
            subscript_chain.append(("getattr", attr_name))
            pos_start = pos_end
        elif start_char == "[":
            pos_end = find_pair(suffix, pos_start)
            if pos_end == -1:
                pos_end = len_suffix

            index = int(suffix[pos_start+1:pos_end])
            subscript_chain.append(("getitem", index))
            pos_start = pos_end + 1

    return subscript_chain

def rfind_pair(content, i):
    len_content = len(content)
    if i < 0 or i >= len_content or content[i] not in ")]}":
        return -1

    start_pair = content[i]
    end_pair = None
    if content[i] == ')':
        end_pair = '('
    elif content[i] == ']':
        end_pair = '['
    elif content[i] == '}':
        end_pair = '{'

    n_start_pair = 0
    str_index, comment_index = get_invalid_index(content)
    while i < len_content:
        if i not in str_index and i not in comment_index:
            if content[i] == start_pair:
                n_start_pair += 1
            elif content[i] == end_pair:
                n_start_pair -= 1

            if n_start_pair == 0:
                break

        i -= 1

    return i

__str_index_dict = {}
__comment_index_dict = {}
def get_invalid_index(content):
    global __str_index_dict
    if content in __str_index_dict:
        return __str_index_dict[content], __comment_index_dict[content]

    str_index = set()
    comment_index = set()
    in_comment = False
    in_str = False
    str_start_char = None
    last_is_slash = False
    i = 0
    len_content = len(content)
    while i < len_content:
        should_add = True
        if content[i] == "'" and not last_is_slash:
            if not in_str:
                should_add = False
                in_str = True
                str_start_char = "'"
            elif str_start_char == "'":
                in_str = False
                str_start_char = None
        elif content[i] == '"' and not last_is_slash:
            if not in_str:
                should_add = False
                in_str = True
                str_start_char = '"'
            elif str_start_char == '"':
                in_str = False
                str_start_char = None
        elif content[i] in "#\\" and not in_str:
            in_comment = True
        elif content[i] == "\n" and in_comment:
            in_comment = False
        
        if content[i] == '\\' and in_str:
            last_is_slash = (not last_is_slash)
        else:
            last_is_slash = False

        if should_add:
            if in_str:
                str_index.add(i)
            if in_comment:
                comment_index.add(i)

        i += 1

    __str_index_dict[content] = str_index
    __comment_index_dict[content] = comment_index

    return str_index, comment_index


def delete_python_comments(content):
    i = 0

    str_index, _ = get_invalid_index(content)
    len_content = len(content)

    while i < len_content:
        if i not in str_index:
            if content[i] == "#":
                pos_endl = content.find("\n", i)
                if pos_endl == -1:
                    pos_endl = len_content
                content = content[:i] + content[pos_endl:]
            elif content[i] == '\\':
                pos_endl = content.find("\n", i)
                if pos_endl == -1:
                    pos_endl = len_content
                else:
                    pos_endl += 1

                content = content[:i] + content[pos_endl:]

        i += 1

    return content

def split_var_str(var_str):
    pos = len(var_str)-1
    if var_str[pos] in ")]}":
        while pos >= 0 and var_str[pos] in ")]}":
            pos = rfind_pair(var_str, pos)
            pos -= 1
            pos = rskip_space(var_str, pos)
        return var_str[:pos+1], var_str[pos+1:]
    else:
        pos = var_str.rfind(".")
        if pos != -1:
            return var_str[:pos], var_str[pos:]
        else:
            return "", ""

def di(id_var):
    return PyObj_FromPtr(id_var)

def LP_LP_c_char(str_list):
    lp_c_char = ctypes.POINTER(ctypes.c_char)
    len_str_list = len(str_list)
    result = (lp_c_char*len_str_list)()
    for i in range(len_str_list):
        result[i] = ctypes.create_string_buffer(str_list[i].encode("utf-8"))
    return result

def capacity_of(length):
    if length <= 0:
        return 2
    else:
        return int(math.pow(2, math.floor(1 + math.log2(length))))

@checktype
def quat_to_mat4(q:glm.quat):
    w = q.w
    x = q.x
    y = q.y
    z = q.z

    return glm.mat4(1-2*(y**2 + z**2), 2*(x*y + w*z), 2*(x*z - w*y), 0,
                     2*(x*y - w*z), 1-2*(x**2 + z**2), 2*(y*z + w*x), 0,
                     2*(x*z + w*y), 2*(y*z - w*x), 1-2*(x**2 + y**2), 0,
                     0, 0, 0, 1)

@checktype
def quat_to_mat3(q:glm.quat):
    w = q.w
    x = q.x
    y = q.y
    z = q.z

    return glm.mat3(1-2*(y**2 + z**2), 2*(x*y + w*z), 2*(x*z - w*y),
                     2*(x*y - w*z), 1-2*(x**2 + z**2), 2*(y*z + w*x),
                     2*(x*z + w*y), 2*(y*z - w*x), 1-2*(x**2 + y**2))

def scale_to_mat4(s:glm.vec3):
    return glm.mat4(s.x, 0  , 0  , 0,
                    0  , s.y, 0  , 0,
                    0  , 0  , s.z, 0,
                    0  , 0  , 0  , 1)

def scale_to_mat3(s:glm.vec3):
    return glm.mat3(s.x, 0  , 0  ,
                    0  , s.y, 0  ,
                    0  , 0  , s.z)

def translate_to_mat4(t:glm.vec3):
    return glm.mat4(1  , 0  , 0  , 0,
                    0  , 1  , 0  , 0,
                    0  , 0  , 1  , 0,
                    t.x, t.y, t.z, 1)


from .FBOAttachment import FBOAttachment
from .GLInfo import GLInfo
from .utils import checktype
from .helper import get_dtype

from OpenGL import GL

class RBO(FBOAttachment):
	
	_basic_info = \
	{
		"gen_func": GL.glGenRenderbuffers,
		"bind_func": GL.glBindRenderbuffer,
		"del_func": GL.glDeleteRenderbuffers,
		"target_type": GL.GL_RENDERBUFFER,
		"binding_type": GL.GL_RENDERBUFFER_BINDING,
		"need_number": True,
	}

	@checktype
	def __init__(self, width:int=0, height:int=0, samples:int=None, internal_format:GLInfo.internal_formats=GL.GL_DEPTH24_STENCIL8):
		FBOAttachment.__init__(self)

		self._samples = samples
		self._width = width
		self._height = height
		self._internal_format = internal_format
		self._param_changed = True

	def bind(self, update_fbo:bool=False, force_update_image:bool=False):
		FBOAttachment.bind(self, update_fbo, force_update_image)

		if self._param_changed or force_update_image:
			if self._samples is not None:
				GL.glRenderbufferStorageMultisample(GL.GL_RENDERBUFFER, self._samples, self._internal_format, self._width, self._height)
			else:
				GL.glRenderbufferStorage(GL.GL_RENDERBUFFER, self._internal_format, self._width, self._height)
			self._param_changed = False

	@property
	def samples(self):
		return self._samples
	
	@samples.setter
	def samples(self, samples:int):
		if self._samples != samples:
			self._samples = samples
			self._param_changed = True

	@property
	def dtype(self):
		return get_dtype(self._internal_format)

	@checktype
	def malloc(self, width:int, height:int, samples:int=None, layers:int=None, internal_format:GLInfo.internal_formats=None):
		self.width = width
		self.height = height

		if internal_format is None:
			internal_format = internal_format

		if internal_format is None:
			internal_format = GL.GL_DEPTH24_STENCIL8

		self.internal_format = internal_format

		if samples is not None:
			self.samples = samples

	@property
	def width(self):
		return self._width

	@width.setter
	def width(self, width):
		if self._width != width:
			self._width = width
			self._param_changed = True

	@property
	def height(self):
		return self._height

	@height.setter
	def height(self, height):
		if self._height != height:
			self._height = height
			self._param_changed = True

	@property
	def internal_format(self):
		return self._internal_format

	@internal_format.setter
	@checktype
	def internal_format(self, format:GLInfo.internal_formats):
		if self._internal_format != format:
			self._internal_format = format
			self._param_changed = True



from ..Mesh import Mesh

from glass.utils import checktype
from glass import Vertex

import glm
import math

def init_Dodecahedron(cls):
    phi1 = (math.sqrt(5)-1)/2
    phi2 = (math.sqrt(5)+1)/2
    cls.base_positions = \
    [
        glm.normalize(glm.vec3(-1, -1, -1)), # 0
        glm.normalize(glm.vec3( 1, -1, -1)), # 1
        glm.normalize(glm.vec3( 1,  1, -1)), # 2
        glm.normalize(glm.vec3(-1,  1, -1)), # 3

        glm.normalize(glm.vec3(-1, -1,  1)), # 4
        glm.normalize(glm.vec3( 1, -1,  1)), # 5
        glm.normalize(glm.vec3( 1,  1,  1)), # 6
        glm.normalize(glm.vec3(-1,  1,  1)), # 7

        glm.normalize(glm.vec3(0, -phi2, -phi1)), # 8
        glm.normalize(glm.vec3(0,  phi2, -phi1)), # 9
        glm.normalize(glm.vec3(0,  phi2,  phi1)), # 10
        glm.normalize(glm.vec3(0, -phi2,  phi1)), # 11

        glm.normalize(glm.vec3(-phi1, 0, -phi2)), # 12
        glm.normalize(glm.vec3(phi1, 0, -phi2)),  # 13
        glm.normalize(glm.vec3(phi1, 0, phi2)),   # 14
        glm.normalize(glm.vec3(-phi1, 0, phi2)),  # 15

        glm.normalize(glm.vec3(-phi2, -phi1, 0)), # 16
        glm.normalize(glm.vec3(phi2, -phi1, 0)),  # 17
        glm.normalize(glm.vec3(phi2, phi1, 0)),   # 18
        glm.normalize(glm.vec3(-phi2, phi1, 0)),  # 19
    ]

    cls.base_indices = \
    [
        # 0
        glm.uvec3(4, 11, 5),
        glm.uvec3(4, 5, 14),
        glm.uvec3(4, 14, 15),

        # 1
        glm.uvec3(8, 1, 17),
        glm.uvec3(8, 17, 5),
        glm.uvec3(8, 5, 11),

        # 2
        glm.uvec3(0, 8, 11),
        glm.uvec3(0, 11, 4),
        glm.uvec3(0, 4, 16),

        # 3
        glm.uvec3(17, 18, 6),
        glm.uvec3(17, 6, 14),
        glm.uvec3(17, 14, 5),

        # 4
        glm.uvec3(13, 2, 18),
        glm.uvec3(13, 18, 17),
        glm.uvec3(13, 17, 1),

        # 5
        glm.uvec3(12, 13, 1),
        glm.uvec3(12, 1, 8),
        glm.uvec3(12, 8, 0),

        # 6
        glm.uvec3(3, 12, 0),
        glm.uvec3(3, 0, 16),
        glm.uvec3(3, 16, 19),

        # 7
        glm.uvec3(19, 16, 4),
        glm.uvec3(19, 4, 15), 
        glm.uvec3(19, 15, 7),

        # 8
        glm.uvec3(6, 10, 7),
        glm.uvec3(6, 7, 15),
        glm.uvec3(6, 15, 14),

        # 9
        glm.uvec3(2, 9, 10),
        glm.uvec3(2, 10, 6),
        glm.uvec3(2, 6, 18),

        # 10
        glm.uvec3(13, 12, 3),
        glm.uvec3(13, 3, 9),
        glm.uvec3(13, 9, 2),

        # 11
        glm.uvec3(9, 3, 19),
        glm.uvec3(9, 19, 7),
        glm.uvec3(9, 7, 10)
    ]

    cls.edge_length = 2*phi1/math.sqrt(3)

    cls.fixed_vertices = []
    theta = 2/5*math.pi
    for i, index in enumerate(cls.base_indices):
        pos0 = cls.base_positions[index[0]]
        pos1 = cls.base_positions[index[1]]
        pos2 = cls.base_positions[index[2]]

        tex_coord0 = None
        tex_coord1 = None
        tex_coord2 = None
        if i % 3 == 0:
            tex_coord0 = 0.5*glm.vec3(1 - math.sin(2*theta), 1 + math.cos(2*theta), 0)
            tex_coord1 = 0.5*glm.vec3(1 + math.sin(2*theta), 1 + math.cos(2*theta), 0)
            tex_coord2 = 0.5*glm.vec3(1 + math.sin(theta), 1 + math.cos(theta), 0)
        elif i % 3 == 1:
            tex_coord0 = 0.5*glm.vec3(1 - math.sin(2*theta), 1 + math.cos(2*theta), 0)
            tex_coord1 = 0.5*glm.vec3(1 + math.sin(theta), 1 + math.cos(theta), 0)
            tex_coord2 = glm.vec3(0.5, 1, 0)
        elif i % 3 == 2:
            tex_coord0 = 0.5*glm.vec3(1 - math.sin(2*theta), 1 + math.cos(2*theta), 0)
            tex_coord1 = glm.vec3(0.5, 1, 0)
            tex_coord2 = 0.5*glm.vec3(1-math.sin(theta), 1+math.cos(theta), 0)
            
        v1 = pos1 - pos0
        v2 = pos2 - pos0
        normal = glm.normalize(glm.cross(v1, v2))

        vertex1 = Vertex(
            position=pos0,
            normal=normal,
            tex_coord=tex_coord0
        )
        vertex2 = Vertex(
            position=pos1,
            normal=normal,
            tex_coord=tex_coord1
        )
        vertex3 = Vertex(
            position=pos2,
            normal=normal,
            tex_coord=tex_coord2
        )
        cls.fixed_vertices.append(vertex1)
        cls.fixed_vertices.append(vertex2)
        cls.fixed_vertices.append(vertex3)

    return cls

@init_Dodecahedron
class Dodecahedron(Mesh):

    @checktype
    def __init__(self, radius=1,
                 color:(glm.vec3,glm.vec4)=glm.vec4(0.396, 0.74151, 0.69102, 1), back_color:(glm.vec3,glm.vec4)=None,
                 stable=False, normalize_tex_coord=False,
                 name:str=""):
        Mesh.__init__(self, color=color, back_color=back_color, name=name, block=True)
        self.__radius = radius
        self.__stable = stable
        self.__normalize_tex_coord = normalize_tex_coord
        self.start_building()

    def build(self):
        self.is_closed = True
        self.self_calculated_normal = True

        vertices = self.vertices
        indices = self.indices
        radius = self.__radius
        stable = self.__stable
        normalize_tex_coord = self.__normalize_tex_coord

        quat = glm.quat(1, 0, 0, 0)
        if stable:
            half_theta = 0.25 * math.asin(2/5*math.sqrt(5))
            quat = glm.quat(math.cos(half_theta), math.sin(half_theta), 0, 0)

        sin_pi_5 = math.sin(math.pi/5)
        for i, fix_vert in enumerate(Dodecahedron.fixed_vertices):
            tex_coord = fix_vert.tex_coord
            if not normalize_tex_coord:
                tex_coord = radius*(tex_coord - glm.vec3(0.5, 0.5, 0))*Dodecahedron.edge_length/sin_pi_5 + glm.vec3(0.5, 0.5, 0)
            
            vertices[i] = Vertex(
                position = radius * (quat * fix_vert.position),
                normal = quat * fix_vert.normal,
                tex_coord = tex_coord
            )

        n_vertices = len(Dodecahedron.fixed_vertices)
        for i in range(0, n_vertices, 3):
            indices[int(i/3)] = glm.uvec3(i, i+1, i+2)
            self.generate_temp_TBN(vertices[i], vertices[i+1], vertices[i+2])

        del vertices[n_vertices:]
        del indices[int(n_vertices/3):]

    @property
    def radius(self):
        return self.__radius
    
    @radius.setter
    @Mesh.param_setter
    def radius(self, radius:float):
        self.__radius = radius

    @property
    def normalize_tex_coord(self):
        return self.__normalize_tex_coord
    
    @normalize_tex_coord.setter
    @Mesh.param_setter
    def normalize_tex_coord(self, flag:bool):
        self.__normalize_tex_coord = flag

    @property
    def is_stable(self):
        return self.__stable
    
    @is_stable.setter
    @Mesh.param_setter
    def is_stable(self, flag:bool):
        self.__stable = flag



from ..Mesh import Mesh

from glass.utils import checktype
from glass import Vertex

import glm
import math

class ConeTrustum(Mesh):

    @checktype
    def __init__(self, bottom_radius:float=2, top_radius:float=1, height:float=1,
                 start_angle:float=0, span_angle:float=360, n_divide:int=100,
                 color:(glm.vec3,glm.vec4)=glm.vec4(0.396, 0.74151, 0.69102, 1), back_color:(glm.vec3,glm.vec4)=None,
                 normalize_tex_coord=False,
                 name:str="", block=True):
        Mesh.__init__(self, color=color, back_color=back_color, name=name, block=block)
        self.__bottom_radius = bottom_radius
        self.__top_radius = top_radius
        self.__height = height
        self.__start_angle = start_angle
        self.__span_angle = span_angle
        self.__n_divide = n_divide
        self.__normalize_tex_coord = normalize_tex_coord
        self.start_building()

    def build(self):
        self.is_closed = True
        self.self_calculated_normal = True

        vertices = self.vertices
        indices = self.indices

        bottom_radius = self.__bottom_radius
        top_radius = self.__top_radius
        height = self.__height
        start_angle = self.__start_angle/180*math.pi
        span_angle = self.__span_angle/180*math.pi
        n_divide = self.__n_divide
        normalize_tex_coord = self.__normalize_tex_coord

        i_vertex = 0
        i_index = 0

        # 圆台顶面中心点
        vertex_top_center = Vertex()
        vertex_top_center.position = glm.vec3(0, 0, height)
        vertex_top_center.normal = glm.vec3(0, 0, 1)
        vertex_top_center.tex_coord = glm.vec3(0.5, 0.5, 0)

        vertices[i_vertex] = vertex_top_center
        i_vertex += 1

        # 圆台底面中心点
        vertex_bottom_center = Vertex()
        vertex_bottom_center.position = glm.vec3(0)
        vertex_bottom_center.normal = glm.vec3(0, 0, -1)
        vertex_bottom_center.tex_coord = glm.vec3(0.5, 0.5, 0)

        vertices[i_vertex] = vertex_bottom_center
        i_vertex += 1

        k = math.sqrt(1 + (height/(bottom_radius - top_radius))**2)
        tex_coord_top_side_radius = 0.5/bottom_radius*top_radius
        if not normalize_tex_coord:
            tex_coord_top_side_radius = top_radius * k
        
        tex_coord_bottom_side_radius = 0.5
        if not normalize_tex_coord:
            tex_coord_bottom_side_radius = bottom_radius * k

        tex_coord_bottom_radius = 0.5
        if not normalize_tex_coord:
            tex_coord_bottom_radius = bottom_radius

        tex_coord_top_radius = 0.5/bottom_radius*top_radius
        if not normalize_tex_coord:
            tex_coord_top_radius = top_radius

        for j in range(n_divide):
            theta = start_angle + span_angle * j/(n_divide-1)
            cos_theta = math.cos(theta)
            sin_theta = math.sin(theta)

            top = glm.vec3(top_radius * cos_theta,
                           top_radius * sin_theta,
                           height)
            bottom = glm.vec3(bottom_radius * cos_theta,
                                   bottom_radius * sin_theta,
                                   0)

            to_right = glm.vec3(-sin_theta, cos_theta, 0)
            to_top = top - bottom
            normal = glm.normalize(glm.cross(to_right, to_top))

            vertex_top_top = Vertex()
            vertex_top_top.position = top
            vertex_top_top.normal = glm.vec3(0, 0, 1)
            vertex_top_top.tex_coord = glm.vec3(0.5+tex_coord_top_radius*cos_theta, 0.5+tex_coord_top_radius*sin_theta, 0)

            vertex_top_side = Vertex()
            vertex_top_side.position = top
            vertex_top_side.normal = normal
            vertex_top_side.tex_coord = glm.vec3(0.5+tex_coord_top_side_radius*cos_theta, 0.5+tex_coord_top_side_radius*sin_theta, 0)

            vertex_bottom_side = Vertex()
            vertex_bottom_side.position = bottom
            vertex_bottom_side.normal = normal
            vertex_bottom_side.tex_coord = glm.vec3(0.5+tex_coord_bottom_side_radius*cos_theta, 0.5+tex_coord_bottom_side_radius*sin_theta, 0)

            vertex_bottom_bottom = Vertex()
            vertex_bottom_bottom.position = bottom
            vertex_bottom_bottom.normal = glm.vec3(0, 0, -1)
            vertex_bottom_bottom.tex_coord = glm.vec3(0.5+tex_coord_bottom_radius*cos_theta, 0.5+tex_coord_bottom_radius*sin_theta, 0)

            vertices[i_vertex] = vertex_top_top # 2 + 4*j
            i_vertex += 1

            vertices[i_vertex] = vertex_top_side # 2 + 4*j + 1
            i_vertex += 1

            vertices[i_vertex] = vertex_bottom_side # 2 + 4*j + 2
            i_vertex += 1

            vertices[i_vertex] = vertex_bottom_bottom # 2 + 4*j + 3
            i_vertex += 1

            if j > 0:
                # 圆台顶面
                triangle = glm.uvec3(0, 0, 0)
                triangle[0] = 2 + 4*j
                triangle[1] = 0
                triangle[2] = 2 + 4*j - 4
                indices[i_index] = triangle
                i_index += 1
                self.generate_temp_TBN(vertices[triangle[0]], vertices[triangle[1]], vertices[triangle[2]])

                # 圆台侧面 1
                triangle = glm.uvec3(0, 0, 0)
                triangle[0] = 2 + 4*j + 2
                triangle[1] = 2 + 4*j + 1
                triangle[2] = 2 + 4*j - 3
                indices[i_index] = triangle
                i_index += 1
                self.generate_temp_TBN(vertices[triangle[0]], vertices[triangle[1]], vertices[triangle[2]])

                # 圆台侧面 2
                triangle = glm.uvec3(0, 0, 0)
                triangle[0] = 2 + 4*j + 2
                triangle[1] = 2 + 4*j - 3
                triangle[2] = 2 + 4*j - 2
                indices[i_index] = triangle
                i_index += 1
                self.generate_temp_TBN(vertices[triangle[0]], vertices[triangle[1]], vertices[triangle[2]])

                # 圆台底面
                triangle = glm.uvec3(0, 0, 0)
                triangle[0] = 2 + 4*j + 3
                triangle[1] = 2 + 4*j - 1
                triangle[2] = 1
                indices[i_index] = triangle
                i_index += 1
                self.generate_temp_TBN(vertices[triangle[0]], vertices[triangle[1]], vertices[triangle[2]])
            
                yield

        del vertices[i_vertex:]
        del indices[i_index:]

    @property
    def n_divide(self):
        return self.__n_divide
    
    @n_divide.setter
    @Mesh.param_setter
    def n_divide(self, n_divide:int):
        self.__n_divide = n_divide

    @property
    def start_angle(self):
        return self.__start_angle
    
    @start_angle.setter
    @Mesh.param_setter
    def start_angle(self, angle:float):
        self.__start_angle = angle

    @property
    def span_angle(self):
        return self.__span_angle
    
    @span_angle.setter
    @Mesh.param_setter
    def span_angle(self, angle:float):
        self.__span_angle = angle

    @property
    def bottom_radius(self):
        return self.__bottom_radius
    
    @bottom_radius.setter
    @Mesh.param_setter
    def bottom_radius(self, bottom_radius:float):
        self.__bottom_radius = bottom_radius

    @property
    def top_radius(self):
        return self.__top_radius
    
    @top_radius.setter
    @Mesh.param_setter
    def top_radius(self, top_radius:float):
        self.__top_radius = top_radius

    @property
    def height(self):
        return self.__height
    
    @height.setter
    @Mesh.param_setter
    def height(self, height:float):
        self.__height = height

    @property
    def normalize_tex_coord(self):
        return self.__normalize_tex_coord
    
    @normalize_tex_coord.setter
    @Mesh.param_setter
    def normalize_tex_coord(self, flag:bool):
        self.__normalize_tex_coord = flag


import wget
import hashlib
import os
import time
import requests

def md5(file_name):
    if not os.path.isfile(file_name):
        return ""

    md5_hash = hashlib.md5()
    content = open(file_name, "rb").read()
    md5_hash.update(content)

    return md5_hash.hexdigest()

def download(url, target_file, md5_str:str=""):
    target_folder = os.path.dirname(os.path.abspath(target_file))
    if not os.path.isdir(target_folder):
        os.makedirs(target_folder)

    if md5_str:
        current_md5 = md5(target_file)
        if current_md5 == md5_str:
            return

    times = 0
    while True:
        if os.path.isfile(target_file):
            os.remove(target_file)

        print(f"downloading {os.path.basename(target_file)}...")
        try:
            wget.download(url, target_file)
        except:
            respond = requests.get(url)
            if respond.status_code != 200:
                raise RuntimeError(respond.content)
            
            out_file = open(target_file, "wb")
            out_file.write(respond.content)
            out_file.close()

        if md5_str:
            current_md5 = md5(target_file)
            if current_md5 == md5_str:
                return
        else:
            if os.path.isfile(target_file):
                return

        time.sleep(1)

        times += 1
        if times > 4:
            raise RuntimeError(f"download {url} to {target_file} retry over times")
        else:
            print("download failed, retry...")



from ..Mesh import Mesh

from glass.utils import checktype
from glass import Vertex

import glm
import math

def init_Icosahedron(cls):
    phi = (1+math.sqrt(5))/2
    cls.base_positions = \
    [
        glm.normalize(glm.vec3(-1,    0,    phi)), # 0
        glm.normalize(glm.vec3( 1,    0,    phi)), # 1
        glm.normalize(glm.vec3(-1,    0,   -phi)), # 2
        glm.normalize(glm.vec3( 1,    0,   -phi)), # 3
        glm.normalize(glm.vec3( 0,    phi,  1  )), # 4
        glm.normalize(glm.vec3( 0,    phi, -1  )), # 5
        glm.normalize(glm.vec3( 0,   -phi,  1  )), # 6
        glm.normalize(glm.vec3( 0,   -phi, -1  )), # 7
        glm.normalize(glm.vec3( phi,  1,    0  )), # 8
        glm.normalize(glm.vec3(-phi,  1,    0  )), # 9
        glm.normalize(glm.vec3( phi, -1,    0  )), # 10
        glm.normalize(glm.vec3(-phi, -1,    0  ))  # 11
    ]

    cls.base_indices = \
    [
        glm.uvec3(1 , 4 , 0 ),
        glm.uvec3(4 , 9 , 0 ),
        glm.uvec3(4 , 5 , 9 ),
        glm.uvec3(8 , 5 , 4 ),
        glm.uvec3(1 , 8 , 4 ),
        glm.uvec3(1 , 10, 8 ),
        glm.uvec3(10, 3 , 8 ),
        glm.uvec3(8 , 3 , 5 ),
        glm.uvec3(3 , 2 , 5 ),
        glm.uvec3(3 , 7 , 2 ),
        glm.uvec3(3 , 10, 7 ),
        glm.uvec3(10, 6 , 7 ),
        glm.uvec3(6 , 11, 7 ),
        glm.uvec3(6 , 0 , 11),
        glm.uvec3(6 , 1 , 0 ),
        glm.uvec3(10, 1 , 6 ),
        glm.uvec3(11, 0 , 9 ),
        glm.uvec3(2 , 11, 9 ),
        glm.uvec3(5 , 2 , 9 ),
        glm.uvec3(11, 2 , 7 )
    ]

    cls.edge_length = 2/math.sqrt(1 + phi*phi)

    cls.fixed_vertices = []
    for index in cls.base_indices:
        pos0 = cls.base_positions[index[0]]
        pos1 = cls.base_positions[index[1]]
        pos2 = cls.base_positions[index[2]]

        v1 = pos1 - pos0
        v2 = pos2 - pos0
        normal = glm.normalize(glm.cross(v1, v2))

        vertex1 = Vertex(
            position=pos0,
            normal=normal,
            tex_coord=glm.vec3(0.5, 1, 0)
        )
        vertex2 = Vertex(
            position=pos1,
            normal=normal,
            tex_coord=glm.vec3(0.5-math.sqrt(3)/4, 0.25, 0)
        )
        vertex3 = Vertex(
            position=pos2,
            normal=normal,
            tex_coord=glm.vec3(0.5+math.sqrt(3)/4, 0.25, 0)
        )
        cls.fixed_vertices.append(vertex1)
        cls.fixed_vertices.append(vertex2)
        cls.fixed_vertices.append(vertex3)

    return cls

@init_Icosahedron
class Icosahedron(Mesh):

    @checktype
    def __init__(self, radius=1,
                 color:(glm.vec3,glm.vec4)=glm.vec4(0.396, 0.74151, 0.69102, 1), back_color:(glm.vec3,glm.vec4)=None,
                 stable=False, normalize_tex_coord=False,
                 name:str=""):
        Mesh.__init__(self, color=color, back_color=back_color, name=name, block=True)
        self.__radius = radius
        self.__stable = stable
        self.__normalize_tex_coord = normalize_tex_coord
        self.start_building()

    def build(self):
        self.is_closed = True
        self.self_calculated_normal = True

        vertices = self.vertices
        indices = self.indices
        radius = self.__radius
        stable = self.__stable
        normalize_tex_coord = self.__normalize_tex_coord

        quat = glm.quat(1, 0, 0, 0)
        if stable:
            half_theta = 0.25 * math.asin(2/3)
            quat = glm.quat(math.cos(half_theta), math.sin(half_theta), 0, 0)

        for i, fix_vert in enumerate(Icosahedron.fixed_vertices):
            tex_coord = fix_vert.tex_coord
            if not normalize_tex_coord:
                tex_coord = 2*radius*(tex_coord - glm.vec3(0.5, 0.5, 0))*Icosahedron.edge_length/math.sqrt(3) + glm.vec3(0.5, 0.5, 0)
            
            vertices[i] = Vertex(
                position = radius * (quat * fix_vert.position),
                normal = quat * fix_vert.normal,
                tex_coord = tex_coord
            )

        n_vertices = len(Icosahedron.fixed_vertices)
        for i in range(0, n_vertices, 3):
            indices[int(i/3)] = glm.uvec3(i, i+1, i+2)
            self.generate_temp_TBN(vertices[i], vertices[i+1], vertices[i+2])

        del vertices[n_vertices:]
        del indices[int(n_vertices/3):]

    @property
    def radius(self):
        return self.__radius
    
    @radius.setter
    @Mesh.param_setter
    def radius(self, radius:float):
        self.__radius = radius

    @property
    def normalize_tex_coord(self):
        return self.__normalize_tex_coord
    
    @normalize_tex_coord.setter
    @Mesh.param_setter
    def normalize_tex_coord(self, flag:bool):
        self.__normalize_tex_coord = flag

    @property
    def is_stable(self):
        return self.__stable
    
    @is_stable.setter
    @Mesh.param_setter
    def is_stable(self, flag:bool):
        self.__stable = flag




from ..Mesh import Mesh

from glass.utils import checktype
from glass import Vertex

import glm
import math

def init_Tetrahedron(cls):
    sqrt_6 = math.sqrt(6)
    sqrt_2 = math.sqrt(2)
    sqrt_3 = math.sqrt(3)
    cls.edge_length = 2/3*sqrt_6
    cls.base_positions = \
    [
        glm.normalize(glm.vec3( 2/3*sqrt_2, 0, -1/3)), # 0
        glm.normalize(glm.vec3( -sqrt_2/3, sqrt_6/3, -1/3)), # 1
        glm.normalize(glm.vec3( -sqrt_2/3,   -sqrt_6/3, -1/3)), # 2
        glm.normalize(glm.vec3( 0,    0,   1)), # 3
    ]

    cls.base_indices = \
    [
        glm.uvec3(0 , 2 , 1),
        glm.uvec3(3 , 0 , 1 ),
        glm.uvec3(3 , 1 , 2 ),
        glm.uvec3(3 , 2 , 0 )
    ]

    cls.fixed_vertices = []
    for index in cls.base_indices:
        v1 = cls.base_positions[index[1]] - cls.base_positions[index[0]]
        v2 = cls.base_positions[index[2]] - cls.base_positions[index[0]]
        normal = glm.normalize(glm.cross(v1, v2))

        vertex1 = Vertex(
            position=cls.base_positions[index[0]],
            normal=normal,
            tex_coord=glm.vec3(0.5, 1, 0)
        )
        vertex2 = Vertex(
            position=cls.base_positions[index[1]],
            normal=normal,
            tex_coord=glm.vec3(0.5-sqrt_3/4, 0.25, 0)
        )
        vertex3 = Vertex(
            position=cls.base_positions[index[2]],
            normal=normal,
            tex_coord=glm.vec3(0.5+sqrt_3/4, 0.25, 0)
        )
        cls.fixed_vertices.append(vertex1)
        cls.fixed_vertices.append(vertex2)
        cls.fixed_vertices.append(vertex3)

    return cls

@init_Tetrahedron
class Tetrahedron(Mesh):

    @checktype
    def __init__(self, radius=1,
                 color:(glm.vec3,glm.vec4)=glm.vec4(0.396, 0.74151, 0.69102, 1), back_color:(glm.vec3,glm.vec4)=None,
                 normalize_tex_coord=False,
                 name:str=""):
        Mesh.__init__(self, color=color, back_color=back_color, name=name, block=True)
        self.__radius = radius
        self.__normalize_tex_coord = normalize_tex_coord
        self.start_building()

    def build(self):
        self.is_closed = True
        vertices = self.vertices
        indices = self.indices
        radius = self.__radius
        normalize_tex_coord = self.__normalize_tex_coord

        for i, fix_vert in enumerate(Tetrahedron.fixed_vertices):
            tex_coord = fix_vert.tex_coord
            if not normalize_tex_coord:
                tex_coord = 2*radius*(tex_coord - glm.vec3(0.5, 0.5, 0))*Tetrahedron.edge_length/math.sqrt(3) + glm.vec3(0.5, 0.5, 0)
            
            vertices[i] = Vertex(
                position = radius * fix_vert.position,
                normal = fix_vert.normal,
                tex_coord = tex_coord
            )

        for i in range(0, len(Tetrahedron.fixed_vertices), 3):
            indices[int(i/3)] = glm.uvec3(i, i+1, i+2)
            self.generate_temp_TBN(vertices[i], vertices[i+1], vertices[i+2])

    @property
    def radius(self):
        return self.__radius
    
    @radius.setter
    @Mesh.param_setter
    def radius(self, radius:float):
        self.__radius = radius

    @property
    def normalize_tex_coord(self):
        return self.__normalize_tex_coord
    
    @normalize_tex_coord.setter
    @Mesh.param_setter
    def normalize_tex_coord(self, flag:bool):
        self.__normalize_tex_coord = flag



from .Vertices import Vertex, Vertices
from .GLInfo import GLInfo
from .utils import checktype

from OpenGL import GL

class Instance(Vertex):
    def __init__(self, **kwargs):
        Vertex.__init__(self, **kwargs)

class Instances(Vertices):

    element_type = Instance

    @checktype
    def __init__(self, values=None, draw_type:GLInfo.draw_types=GL.GL_STATIC_DRAW):
        self._path_index_map = {}

        if isinstance(values, (list,type(None))):
            Vertices.__init__(self, values=values, draw_type=draw_type)
        else:
            Vertices.__init__(self, draw_type=draw_type)
            self.update(values)

    def __hash__(self):
        return id(self)
    
    def __eq__(self, other):
        return (id(self) == id(other))

    def __setitem__(self, index:(int,slice,str), instance):
        if isinstance(index, str):
            path = index
            if path in self._path_index_map:
                index = self._path_index_map[path]
                Vertices.__setitem__(self, index, instance)
            else:
                self._path_index_map[path] = len(self)
                self.append(instance)
        else:
            Vertices.__setitem__(self, index, instance)

    def __getitem__(self, index:(int,slice,str)):
        if isinstance(index, str):
            path = index
            if path in self._path_index_map:
                index = self._path_index_map[path]
                return Vertices.__getitem__(self, index)
            else:
                raise KeyError(path)
        else:
            return Vertices.__getitem__(self, index)
        
    def __delitem__(self, index:(int,str,slice)):
        if isinstance(index, str):
            path = index
            if path in self._path_index_map:
                index = self._path_index_map[path]
                Vertices.__delitem__(self, index)
                for p in self._path_index_map:
                    i = self._path_index_map[p]
                    if i > index:
                        self._path_index_map[p] = i-1
            else:
                raise KeyError(path)
        elif isinstance(index, int):
            Vertices.__delitem__(self, index)
            for p in self._path_index_map:
                i = self._path_index_map[p]
                if i > index:
                    self._path_index_map[p] = i-1
        elif isinstance(index, slice):
            start, stop, step = self._process_slice(index)
            for i in range(start, stop, step):
                del self[i]

    def pop(self, index:(int,str,slice)):
        if isinstance(index, str):
            path = index
            if path in self._path_index_map:
                index = self._path_index_map[path]
                value = Vertices.pop(self, index)
                for p in self._path_index_map:
                    i = self._path_index_map[p]
                    if i > index:
                        self._path_index_map[p] = i - 1
                return value
            else:
                raise KeyError(path)
        else:
            return Vertices.pop(self, index)
        
    def insert(self, index:int, value):
        Vertices.insert(self, index, value)
        for p in self._path_index_map:
            i = self._path_index_map[p]
            if i >= index:
                self._path_index_map[p] = i + 1
        
    def clear(self):
        Vertices.clear(self)
        self._path_index_map.clear()

    def update(self, values):
        if isinstance(values, list):
            Vertices.update(self, values)
        else:
            for key in values.keys():
                value = values[key]
                if key in self:
                    self[key].update(value)
                else:
                    self[key] = value

    def __contains__(self, key)->bool:
        if isinstance(key, str):
            return (key in self._path_index_map)
        else:
            return Vertices.__contains__(self, key)
        
    def keys(self):
        return self._path_index_map.keys()

    @property
    def divisor(self):
        if not hasattr(self, "_divisor"):
            return 1
        else:
            return self._divisor
    
    @divisor.setter
    @checktype
    def divisor(self, divisor:int):
        if divisor <= 0:
            raise ValueError(f"divisor must be greater than 0, {divisor} were given")
        self._divisor = divisor



from OpenGL import GL
from .utils import di

class Block:

	class Variable:
		def __init__(self, block, name:str)->None:
			self._block_id:int = id(block)
			self._name:str = name
			self._bound_var:object = None
			self._binding_point:int = 0

			atom_info_list = block._block_map[self._name]["atoms"]
			if not atom_info_list:
				return

			self._atom_info_map = {}

			structure_key_list = []
			for atom_info in atom_info_list:
				atom_name = atom_info["name"]
				self._atom_info_map[atom_name] = atom_info
				structure_key_list.append("(" + atom_info["type"] + ", " + atom_name + ")")
			self._structure_key:str = ", ".join(structure_key_list)

		@property
		def block(self):
			return di(self._block_id)

		def bind(self, var:object)->None:
			if var is self._bound_var:
				return

			if var is None:
				self.unbind()
				return

			self.unbind()

			cls = self.block.__class__
			id_var = id(var)
			if id_var not in cls._bound_vars:
				cls._bound_vars[id_var] = {}

			if self._structure_key not in cls._bound_vars[id_var]:
				ssubo = cls.BO()
				ssubo._bound_var = var
				cls._bound_vars[id_var][self._structure_key] = ssubo
			ssubo = cls._bound_vars[id_var][self._structure_key]

			ssubo._atom_info_map = self._atom_info_map
			ssubo._bound_block_vars.add(self)

			self._bound_var = var

		def unbind(self)->None:
			if self._bound_var is None:
				return

			id_var = id(self._bound_var)
			cls = self._block.__class__
			
			if id_var not in cls._bound_vars or \
			   self._structure_key not in cls._bound_vars[id_var]:
				return

			ssubo = cls._bound_vars[id_var][self._structure_key]
			if self in ssubo._bound_block_vars:
				ssubo._bound_block_vars.remove(self)
			if not ssubo._bound_block_vars:
				ssubo.unbind_from_point()

			self._bound_var = None

		def binding_point(self)->int:
			return self._binding_point

		def __del__(self)->None:
			try:
				self.unbind()
			except:
				pass

		def __hash__(self)->int:
			return id(self)
		
		def __eq__(self, other)->bool:
			return (id(self) == id(other))

		def bind_to_point(self, binding_point:int)->None:
			if binding_point < 0:
				return

			if self._binding_point == binding_point:
				return

			block = self.block
			cls_name = block.__class__.__name__
			if cls_name == "UniformBlock":
				GL.glUniformBlockBinding(block.program.id, block._blocks_info[self._name]["index"], binding_point)
			elif cls_name == "ShaderStorageBlock":
				GL.glShaderStorageBlockBinding(block.program.id, block._blocks_info[self._name]["index"], binding_point)

			self._binding_point = binding_point

		def upload(self, force_upload:bool=False)->None:
			if self._bound_var is None:
				return

			cls = self.block.__class__
			id_var = id(self._bound_var)
			if id_var not in cls._bound_vars:
				return

			if self._structure_key not in cls._bound_vars[id_var]:
				return

			ssubo = cls._bound_vars[id_var][self._structure_key]
			assert self in ssubo._bound_block_vars

			if force_upload:
				ssubo.upload(force_upload)
			elif hasattr(self._bound_var, "upload"):
				self._bound_var.upload()

			binding_point = ssubo.bind_to_point(force_bind=False)
			self.bind_to_point(binding_point)

	def __init__(self, program):
		self._program_id = id(program)
		self._block_var_map = {}
		self._auto_upload = True

	@property
	def program(self):
		return di(self._program_id)

	@property
	def auto_upload(self)->bool:
		return self._auto_upload
	
	@auto_upload.setter
	def auto_upload(self, flag:bool)->None:
		self._auto_upload = flag

	def upload(self, force_upload:bool=False)->None:
		for block_var in self._block_var_map.values():
			block_var.upload(force_upload)

	@classmethod
	def upload_var(cls, var:object, force_upload:bool=False)->None:
		id_var = id(var)
		if id_var not in cls._bound_vars:
			return

		for ssubo in cls._bound_vars[id_var].values():
			ssubo.upload(force_upload)

	def __contains__(self, name:str)->bool:
		return (name in self._blocks_info)

	def __getitem__(self, name:str)->Variable:
		if name not in self._blocks_info:
			raise NameError(f"block '{name}' is not defined")

		if name not in self._block_var_map:
			self._block_var_map[name] = Block.Variable(self, name)

		return self._block_var_map[name]
	
	def __setitem__(self, name:str, var:object)->None:
		self[name].bind(var)

	@property
	def _block_map(self):
		if self.__class__.__name__ == "UniformBlock":
			return self.program._uniform_block_map
		elif self.__class__.__name__ == "ShaderStorageBlock":
			return self.program._shader_storage_block_map

	@property
	def _blocks_info(self):
		if self.__class__.__name__ == "UniformBlock":
			return self.program._uniform_blocks_info
		elif self.__class__.__name__ == "ShaderStorageBlock":
			return self.program._shader_storage_blocks_info



from .Mesh import Mesh
from .Camera import Camera

from glass.utils import checktype
from glass import Vertex, samplerCube, ShaderProgram

import glm
from OpenGL import GL
import os

class SkyBox(Mesh):

    @checktype
    def __init__(self, name:str=""):
        Mesh.__init__(self, name=name, block=True)
        self.render_hints.depth_func = "<="
        self.render_hints.cull_face = GL.GL_FRONT
        self.__skybox_map = None
        self.__program = None
        self.should_add_color = False
        self.start_building()

    @property
    def skybox_map(self):
        return self.__skybox_map
    
    @skybox_map.setter
    @checktype
    def skybox_map(self, skybox_map:samplerCube):
        self.__skybox_map = skybox_map
    
    @property
    def is_completed(self):
        return (self.__skybox_map is not None and self.__skybox_map.is_completed)
    
    def __getattr__(self, key):
        if key not in ["right", "left", "top", "bottom", "back", "front", "up", "down"]:
            return super().__getattr__(key)

        if self.skybox_map is None:
            self.skybox_map = samplerCube()

        return self.skybox_map[key]

    def __setattr__(self, key, value):
        if key not in ["right", "left", "top", "bottom", "back", "front", "up", "down"]:
            return super().__setattr__(key, value)

        if self.skybox_map is None:
            self.skybox_map = samplerCube()

        self.skybox_map[key] = value

    def build(self):
        self.self_calculated_normal = True
        self.is_closed = True
        self.x_min, self.x_max = -1, 1
        self.y_min, self.y_max = -1, 1
        self.z_min, self.z_max = -1, 1

        vertices = self.vertices
        indices = self.indices

        # 左面
        vertices[0] = Vertex(position=glm.vec3(-1,  1, -1))
        vertices[1] = Vertex(position=glm.vec3(-1, -1, -1))
        vertices[2] = Vertex(position=glm.vec3(-1, -1,  1))
        vertices[3] = Vertex(position=glm.vec3(-1,  1,  1))

        # 右面
        vertices[4] = Vertex(position=glm.vec3(1, -1, -1))
        vertices[5] = Vertex(position=glm.vec3(1,  1, -1))
        vertices[6] = Vertex(position=glm.vec3(1,  1,  1))
        vertices[7] = Vertex(position=glm.vec3(1, -1,  1))

        # 后面
        vertices[8] = Vertex(position=glm.vec3(-1, -1, -1))
        vertices[9] = Vertex(position=glm.vec3( 1, -1, -1))
        vertices[10] = Vertex(position=glm.vec3( 1, -1,  1))
        vertices[11] = Vertex(position=glm.vec3(-1, -1,  1))

        # 前面
        vertices[12] = Vertex(position=glm.vec3( 1, 1, -1))
        vertices[13] = Vertex(position=glm.vec3(-1, 1, -1))
        vertices[14] = Vertex(position=glm.vec3(-1, 1,  1))
        vertices[15] = Vertex(position=glm.vec3( 1, 1,  1))

        # 下面
        vertices[16] = Vertex(position=glm.vec3(-1,  1, -1))
        vertices[17] = Vertex(position=glm.vec3( 1,  1, -1))
        vertices[18] = Vertex(position=glm.vec3( 1, -1, -1))
        vertices[19] = Vertex(position=glm.vec3(-1, -1, -1))

        # 上面
        vertices[20] = Vertex(position=glm.vec3(-1, -1, 1))
        vertices[21] = Vertex(position=glm.vec3( 1, -1, 1))
        vertices[22] = Vertex(position=glm.vec3( 1,  1, 1))
        vertices[23] = Vertex(position=glm.vec3(-1,  1, 1))

        # 左面
        indices[0] = glm.uvec3(0, 1, 2)
        indices[1] = glm.uvec3(0, 2, 3)

        # 右面
        indices[2] = 4+glm.uvec3(0, 1, 2)
        indices[3] = 4+glm.uvec3(0, 2, 3)

        # 后面
        indices[4] = 8+glm.uvec3(0, 1, 2)
        indices[5] = 8+glm.uvec3(0, 2, 3)

        # 前面
        indices[6] = 12+glm.uvec3(0, 1, 2)
        indices[7] = 12+glm.uvec3(0, 2, 3)
        
        # 下面
        indices[8] = 16+glm.uvec3(0, 1, 2)
        indices[9] = 16+glm.uvec3(0, 2, 3)

        # 上面
        indices[10] = 20+glm.uvec3(0, 1, 2)
        indices[11] = 20+glm.uvec3(0, 2, 3)

    @property
    def program(self):
        if self.__program is None:
            self.__program = ShaderProgram()
            self.__program.compile(os.path.dirname(os.path.abspath(__file__)) + "/glsl/Pipelines/skybox/skybox.vs")
            self.__program.compile(os.path.dirname(os.path.abspath(__file__)) + "/glsl/Pipelines/skybox/skybox.fs")

        return self.__program
    
    def draw(self, camera:Camera):
        scene = camera.scene
        self.program["camera"] = camera
        self.program["skybox_map"] = self.skybox_map
        self.program["sky_distance"] = scene.background.distance
        self.program["fog"] = scene.fog
        Mesh.draw(self, self.program)


from .BO import BO
from .utils import capacity_of

from OpenGL import GL
import numpy as np

class ACBO(BO):

    _basic_info = \
    {
        "gen_func": GL.glGenBuffers,
        "bind_func": GL.glBindBuffer,
        "del_func": GL.glDeleteBuffers,
        "target_type": GL.GL_ATOMIC_COUNTER_BUFFER,
        "binding_type": GL.GL_ATOMIC_COUNTER_BUFFER_BINDING,
        "need_number": True,
    }

    _binding_points_pool = None

    _ACBO_map = {}

    def __init__(self)->None:
        BO.__init__(self)
    
    @staticmethod
    def set(binding:int, offset:int, value:int)->None:
        acbo = None
        if binding not in ACBO._ACBO_map:
            acbo = ACBO()
            acbo.malloc(4 * capacity_of(offset/4), GL.GL_DYNAMIC_COPY)
            ACBO._ACBO_map[binding] = acbo
        else:
            acbo = ACBO._ACBO_map[binding]
            if offset + 4 > acbo.nbytes:
                temp_bo = ACBO()
                temp_bo.malloc(4 * capacity_of(offset/4), GL.GL_DYNAMIC_COPY)
                acbo.copy_to(0, acbo.nbytes, temp_bo, 0)
                acbo.delete()
                acbo = temp_bo
                ACBO._ACBO_map[binding] = acbo

        acbo.bind_to_point(binding)
        acbo.bufferSubData(offset, 4, np.array([int(value)], dtype=np.uint32))

    @staticmethod
    def get(binding:int, offset:int)->int:
        if binding not in ACBO._ACBO_map:
            return 0

        acbo = ACBO._ACBO_map[binding]
        acbo.bind()
        data = np.array([0], np.uint32)
        GL.glGetBufferSubData(GL.GL_ATOMIC_COUNTER_BUFFER, offset, 4, data)
        return int(data[0])

    def bind_to_point(self, binding_point:int)->None:
        self.bind()
        GL.glBindBufferBase(GL.GL_ATOMIC_COUNTER_BUFFER, binding_point, self._id)


class SlideAverageFilter:

    def __init__(self, window_width:int=10)->None:
        self._current_sum = 0
        self._window_width = window_width
        self._data_list = []

    def __call__(self, new_value:(float,int))->float:
        if len(self._data_list) >= self._window_width:
            old_value = self._data_list.pop(0)
            self._current_sum -= old_value

        self._data_list.append(new_value)
        self._current_sum += new_value

        return self._current_sum / len(self._data_list)
    
    @property
    def window_width(self)->int:
        return self._window_width
    
    @window_width.setter
    def window_width(self, window_width:int)->None:
        self._window_width = window_width


from ..Manipulators.Manipulator import Manipulator
from ..Renderers.Renderer import Renderer
from ..PostProcessEffects import PostProcessEffects, BloomEffect, ACESToneMapper, FXAAEffect, DOFEffect, SSAOEffect, ExplosureAdaptor
from ..SlideAverageFilter import SlideAverageFilter
from ..VideoRecorder import VideoRecorder, convert_to_image

from glass import ShaderProgram, GLConfig, GlassConfig, FBO, RBO, sampler2DMS, sampler2D, RenderHints, SSBO, UBO, VAO
from glass.utils import extname, di

import time
import glm
import sys
from OpenGL import GL
import numpy as np
import os
import cv2
import warnings

def __new__(cls, *args, **kwargs):
    if cls.qt.QtWidgets.QApplication.instance() is None:
        cls._app = cls.qt.QtWidgets.QApplication(sys.argv)

    instance = cls.__base__.__new__(cls, *args, **kwargs)
    return instance

def __init__(self, camera, parent=None)->None:
    cls = self.__class__
    cls.__base__.__init__(self, parent=parent)

    self.setFocusPolicy(cls.qt.QtCore.Qt.FocusPolicy.StrongFocus)
    self._samples = 1
    self._samples_set_by_user = False

    self._pressed_keys = set()
    self._last_frame_time = 0
    self._is_cursor_hiden = False
    self._hide_cursor_global_pos = glm.vec2(0)
    self._last_cursor_global_pos = cls.qt.QtCore.QPoint(0, 0)
    self._fps = 0
    self._smooth_fps = 0
    self._before_draw_calls = 0
    self._before_draw_points = 0
    self._before_draw_lines = 0
    self._before_draw_meshes = 0
    self._before_draw_patches = 0
    self._draw_calls = 0
    self._draw_points = 0
    self._draw_lines = 0
    self._draw_meshes = 0
    self._draw_patches = 0
    self._paint_times = 0

    self._fps_filter = SlideAverageFilter()
    self._before_PPE_fbo_ss = None
    self._before_PPE_fbo_ms = None
    self._is_gl_init = False
    self._before_PPE_image = None
    self._video_recorders = []
    
    self._camera_id = id(camera)

    self._manipulator = None
    self._renderer = None
    
    self._render_hints = RenderHints()
    self._render_hints.depth_test = True

    self._listen_cursor_timer = self.startTimer(10)

    self._post_process_effects = PostProcessEffects()

    self._post_process_effects["SSAO"] = SSAOEffect()
    self._post_process_effects["bloom"] = BloomEffect()
    self._post_process_effects["DOF"] = DOFEffect()
    self._post_process_effects["explosure_adaptor"] = ExplosureAdaptor()
    self._post_process_effects["tone_mapper"] = ACESToneMapper()
    self._post_process_effects["FXAA"] = FXAAEffect(internal_format=GL.GL_RGBA8)

    self._post_process_effects["SSAO"].enabled = False
    self._post_process_effects["bloom"].enabled = False
    self._post_process_effects["DOF"].enabled = False
    self._post_process_effects["explosure_adaptor"].enabled = False
    self._post_process_effects["tone_mapper"].enabled = False
    self._post_process_effects["FXAA"].enabled = False

def update(self)->None:
    self._before_PPE_image = None
    self.__class__.__base__.update(self)

def _assign_values_to_PPEs(self):
    self._post_process_effects.camera = self.camera
    self._post_process_effects.depth_map = self.renderer._depth_map
    self._post_process_effects.view_normal_map = self.renderer._view_normal_map
    self._post_process_effects.view_pos_map = self._renderer._view_pos_map

def _update(self)->None:
    self.__class__.__base__.update(self)

def sizeHint(self):
    return self.__class__.qt.QtCore.QSize(800, 600)

@property
def camera(self):
    return di(self._camera_id)

@property
def samples(self)->int:
    return self._samples

@samples.setter
def samples(self, samples:int)->None:
    surface_format = self.__class__.qt.QtGui.QSurfaceFormat()
    surface_format.setSamples(samples)
    self.setFormat(surface_format)
    if not self._is_gl_init:
        self._samples = samples
        self._samples_set_by_user = True

def _set_samples(self, samples:int)->None:
    surface_format = self.__class__.qt.QtGui.QSurfaceFormat()
    surface_format.setSamples(samples)
    self.setFormat(surface_format)
    if not self._is_gl_init:
        self._samples = samples

@property
def renderer(self)->Renderer:
    return self._renderer

@renderer.setter
def renderer(self, renderer:Renderer)->None:
    if self._renderer is renderer:
        return
    
    if renderer is None:
        self._renderer._camera_id = id(None)
        self._renderer = None
        return

    if renderer._camera_id != id(None):
        renderer.camera.screen.renderer = None

    self._renderer = renderer
    renderer._camera_id = self._camera_id
    should_update = renderer.startup()
    if should_update:
        self.update()

@property
def manipulator(self)->Manipulator:
    return self._manipulator

@manipulator.setter
def manipulator(self, manipulator:Manipulator):
    if self._manipulator is manipulator:
        return
    
    if manipulator is None:
        self._manipulator._camera_id = id(None)
        self._manipulator = None
        return

    if manipulator._camera_id != id(None):
        manipulator.camera.screen.manipulator = None

    self._manipulator = manipulator
    manipulator._camera_id = self._camera_id
    should_update = manipulator.startup()
    if should_update:
        self.update()

@property
def render_hints(self)->RenderHints:
    return self._render_hints

@render_hints.setter
def render_hints(self, hint:RenderHints)->None:
    if hint is None:
        hint = RenderHints()

    self._render_hints = hint

def initializeGL(self)->None:
    if GLConfig.major_version < 4:
        raise RuntimeError(f"Current OpenGL version ({GLConfig.version}) is lower than minimum version require (OpenGL 4.3)")

    if GLConfig.minor_version < 3:
        raise RuntimeError(f"Current OpenGL version ({GLConfig.version}) is lower than minimum version require (OpenGL 4.3)")

    if GlassConfig.warning and "GL_ARB_bindless_texture" not in GLConfig.available_extensions:
        warning_message = """
Extension GL_ARB_bindless_texture is not available.
Shadows and dynamic environment mapping will be disabled.
Try to change default graphics card of python.exe to high performance one.
"""
        warnings.warn(warning_message)

    self.makeCurrent()

    if self.camera.scene is None:
        raise RuntimeError(f"{self.camera} is not in any scene, please add it to one scene before show it's screen")

    self._is_gl_init = True

def makeCurrent(self)->None:
    self.__class__.__base__.makeCurrent(self)
    GLConfig.buffered_current_context = GLConfig.current_context
    GLConfig.buffered_viewport = GLConfig.viewport

    SSBO.makeCurrent()
    UBO.makeCurrent()
    VAO.execute_cmd_buffer()

def resizeGL(self, width:int, height:int)->None:
    self.__class__.__base__.resizeGL(self, width, height)
    self._before_PPE_image = None

def _draw_to_before_PPE(self, should_update_scene:bool)->bool:
    if self._before_PPE_image is None or should_update_scene:
        with self._before_PPE_fbo:
            scene = self.camera.scene
            clear_color = scene.fog.apply(scene.background.color, glm.vec3(0), glm.vec3(0,self.camera.far,0))
            with GLConfig.LocalConfig(clear_color=clear_color):
                with self.render_hints:
                    should_update_scene = self.renderer.render() or should_update_scene

        resolved = self._before_PPE_fbo.resolved
        self._before_PPE_image = resolved.color_attachment(0)
        self._post_process_effects.screen_update_time = time.time()

    return should_update_scene

def _mark_draw_calls(self):
    self._before_draw_calls = ShaderProgram.accum_draw_calls()
    self._before_draw_points = ShaderProgram.accum_draw_points()
    self._before_draw_lines = ShaderProgram.accum_draw_lines()
    self._before_draw_meshes = ShaderProgram.accum_draw_meshes()
    self._before_draw_patches = ShaderProgram.accum_draw_patches()

def paintGL(self)->None:
    self.makeCurrent()
    self.frame_started.emit()
    self._mark_draw_calls()

    camera = self.camera
    scene = camera.scene

    should_update_scene = scene.generate_meshes()
    should_update_PPEs = False
    
    if self._video_recorders:
        should_update_scene = self._draw_to_before_PPE(should_update_scene)
        self._assign_values_to_PPEs()
        screen_image = self._post_process_effects.apply(self._before_PPE_image)
        should_update_PPEs = self._post_process_effects.should_update
        screen_image.fbo.draw_to_active(0)

        if self._paint_times > 0:
            view_port = GLConfig.buffered_viewport
            for video_recorder in self._video_recorders:
                video_recorder._start(view_port, self.smooth_fps)
                video_recorder._frame_queue.put((time.time(), screen_image.fbo.data(0)))
    elif self._post_process_effects.has_valid:
        should_update_scene = self._draw_to_before_PPE(should_update_scene)
        self._assign_values_to_PPEs()
        should_update_PPEs = self._post_process_effects.draw_to_active(self._before_PPE_image)
    else:
        clear_color = scene.fog.apply(scene.background.color, glm.vec3(0), glm.vec3(0,self.camera.far,0))
        with GLConfig.LocalConfig(clear_color=clear_color):
            with self.render_hints:
                should_update_scene = self.renderer.render() or should_update_scene

    self._calc_fps()
    self.frame_ended.emit()

    if should_update_scene or should_update_PPEs:
        if should_update_scene:
            self._before_PPE_image = None
            
        self._update()

@property
def _before_PPE_fbo(self)->FBO:
    screen_size = GLConfig.screen_size
    if self.samples > 1:
        if self._before_PPE_fbo_ms is not None:
            self._before_PPE_fbo_ms.resize(screen_size.x, screen_size.y, self.samples)
        else:
            self._before_PPE_fbo_ms = FBO(screen_size.x, screen_size.y, self.samples)
            self._before_PPE_fbo_ms.attach(0, sampler2DMS, GL.GL_RGBA32F)
            self._before_PPE_fbo_ms.attach(GL.GL_DEPTH_ATTACHMENT, RBO)
        return self._before_PPE_fbo_ms
    else:
        if self._before_PPE_fbo_ss is not None:
            self._before_PPE_fbo_ss.resize(screen_size.x, screen_size.y)
        else:
            self._before_PPE_fbo_ss = FBO(screen_size.x, screen_size.y)
            self._before_PPE_fbo_ss.attach(0, sampler2D, GL.GL_RGBA32F)
            self._before_PPE_fbo_ss.attach(GL.GL_DEPTH_ATTACHMENT, RBO)
        return self._before_PPE_fbo_ss
    
def _mouse_parameters(self, mouse_event)->tuple:
    button = mouse_event.button()
    try:
        button_value = button.value
    except:
        button_value = int(button)

    button = Manipulator.MouseButton(button_value)
    try:
        screen_pos = mouse_event.position()
        global_pos = mouse_event.globalPosition()
    except:
        screen_pos = mouse_event.pos()
        global_pos = self.mapToGlobal(screen_pos)

    screen_pos = glm.vec2(screen_pos.x(), screen_pos.y())
    global_pos = glm.vec2(global_pos.x(), global_pos.y())
    return button, screen_pos, global_pos

def _wheel_parameters(self, wheel_event)->tuple:
    angle = glm.vec2(wheel_event.angleDelta().x(), wheel_event.angleDelta().y())
    
    try:
        screen_pos = wheel_event.position()
        global_pos = wheel_event.globalPosition()
    except:
        screen_pos = wheel_event.pos()
        global_pos = self.mapToGlobal(screen_pos)

    screen_pos = glm.vec2(screen_pos.x(), screen_pos.y())
    global_pos = glm.vec2(global_pos.x(), global_pos.y())
    return angle, screen_pos, global_pos

def mousePressEvent(self, mouse_event)->None:
    button, screen_pos, global_pos = self._mouse_parameters(mouse_event)

    self.makeCurrent()

    should_update = False
    if self.manipulator is not None:
        should_update = self.manipulator.on_mouse_pressed(button, screen_pos, global_pos)

    self.mouse_pressed.emit(button, screen_pos, global_pos)

    if should_update:
        self.update()

def mouseReleaseEvent(self, mouse_event)->None:
    button, screen_pos, global_pos = self._mouse_parameters(mouse_event)

    self.makeCurrent()

    should_update = False
    if self.manipulator is not None:
        should_update = self.manipulator.on_mouse_released(button, screen_pos, global_pos)
    
    self.mouse_released.emit(button, screen_pos, global_pos)
    
    if should_update:
        self.update()

def mouseDoubleClickEvent(self, mouse_event)->None:
    button, screen_pos, global_pos = self._mouse_parameters(mouse_event)

    self.makeCurrent()

    should_update = False
    if self.manipulator is not None:
        should_update = self.manipulator.on_mouse_double_clicked(button, screen_pos, global_pos)
    
    self.mouse_double_clicked.emit(button, screen_pos, global_pos)
    
    if should_update:
        self.update()

def _mouseMoveEvent(self, screen_pos:glm.vec2, global_pos:glm.vec2)->bool:
    self.makeCurrent()

    should_update = False
    if self.manipulator is not None:
        should_update = self.manipulator.on_mouse_moved(screen_pos, global_pos)

    self.mouse_moved.emit(screen_pos, global_pos)

    if should_update:
        self.update()

def keyPressEvent(self, key_event)->None:
    key = Manipulator.Key(key_event.key())
    if key not in self._pressed_keys:
        self._pressed_keys.add(key)
    
    self.makeCurrent()

    should_update = False
    if self.manipulator is not None:
        should_update = self.manipulator.on_key_pressed(key)

    self.key_pressed.emit(key)

    if should_update:
        self.update()

def keyReleaseEvent(self, key_event)->None:
    key = Manipulator.Key(key_event.key())
    if key in self._pressed_keys:
        self._pressed_keys.remove(key)

    self.makeCurrent()

    should_update = False
    if self.manipulator is not None:
        should_update = self.manipulator.on_key_released(key)

    self.key_released.emit(key)

    if should_update:
        self.update()

def _keyRepeateEvent(self, keys:set)->bool:
    self.makeCurrent()

    should_update = False
    if self.manipulator is not None:
        should_update = self.manipulator.on_key_repeated(keys)

    self.key_repeated.emit(keys)

    if should_update:
        self.update()

def wheelEvent(self, wheel_event)->None:
    angle, screen_pos, global_pos = self._wheel_parameters(wheel_event)

    self.makeCurrent()

    should_update = False
    if self.manipulator is not None:
        should_update = self.manipulator.on_wheel_scrolled(angle, screen_pos, global_pos)
    
    self.wheel_scrolled.emit(angle, screen_pos, global_pos)

    if should_update:
        self.update()

def _calc_fps(self)->None:
    current_time = time.time()
    dt = current_time - self._last_frame_time
    self._last_frame_time = current_time
    if dt < 1:
        self._fps = 1 / dt
        self._smooth_fps = self._fps_filter(self._fps)

    self._draw_calls = ShaderProgram.accum_draw_calls() - self._before_draw_calls
    self._draw_points = ShaderProgram.accum_draw_points() - self._before_draw_points
    self._draw_lines = ShaderProgram.accum_draw_lines() - self._before_draw_lines
    self._draw_meshes = ShaderProgram.accum_draw_meshes() - self._before_draw_meshes
    self._draw_patches = ShaderProgram.accum_draw_patches() - self._before_draw_patches
    self._paint_times += 1

def hide_cursor(self)->None:
    if self._is_cursor_hiden:
        return
    
    self._is_cursor_hiden = True
    cursor_pos = self.__class__.qt.QtGui.QCursor.pos()
    self._hide_cursor_global_pos = glm.vec2(cursor_pos.x(), cursor_pos.y())
    self.setCursor(self.__class__.qt.QtCore.Qt.CursorShape.BlankCursor)
    return self._hide_cursor_global_pos

def show_cursor(self)->None:
    if not self._is_cursor_hiden:
        return
    
    self._is_cursor_hiden = False
    self.unsetCursor()

def show(self)->None:
    if self.__class__._app is not None and \
        not self.__class__._has_exec:
        self.setWindowTitle("Glass Engine")
        self_folder = os.path.dirname(os.path.abspath(__file__))
        self.setWindowIcon(self.__class__.qt.QtGui.QIcon(self_folder + "/../images/glass_engine_logo64.png"))
        
    self.__class__.__base__.show(self)
    if self.__class__._app is not None and \
        not self.__class__._has_exec:
        self.__class__._app.exec()
        self.__class__._has_exec = True

@property
def post_process_effects(self)->PostProcessEffects:
    return self._post_process_effects

@property
def is_cursor_hiden(self)->bool:
    return self._is_cursor_hiden

@property
def fps(self)->float:
    return self._fps

@property
def smooth_fps(self)->float:
    return self._smooth_fps

@property
def draw_calls(self)->int:
    return self._draw_calls

@property
def draw_points(self)->int:
    return self._draw_points

@property
def draw_lines(self)->int:
    return self._draw_lines

@property
def draw_meshes(self)->int:
    return self._draw_meshes

@property
def draw_patches(self)->int:
    return self._draw_patches

def timerEvent(self, timer_event)->None:
    if timer_event.timerId() == self._listen_cursor_timer:
        cursor_global_pos = self.__class__.qt.QtGui.QCursor.pos()
        last_global_pos = self._last_cursor_global_pos
        self._last_cursor_global_pos = cursor_global_pos

        cursor_global_posF = self.__class__.qt.QtCore.QPointF(cursor_global_pos)

        should_update = False
        if self._pressed_keys:
            should_update = self._keyRepeateEvent(self._pressed_keys) or should_update

        if self._is_cursor_hiden:
            try:
                center_global_posF = self.mapToGlobal(self.__class__.qt.QtCore.QPointF(self.width()/2, self.height()))
            except:
                center = self.__class__.qt.QtCore.QPoint(int(self.width()/2), int(self.height()/2))
                center_global_pos = self.mapToGlobal(center)
                center_global_posF = self.__class__.qt.QtCore.QPointF(float(center_global_pos.x()), float(center_global_pos.y()))
            
            offset = cursor_global_posF - center_global_posF
            buffer = 100
            if offset.x() > buffer or offset.x() < -buffer or \
                offset.y() > buffer or offset.y() < -buffer:
                self._hide_cursor_global_pos.x -= offset.x()
                self._hide_cursor_global_pos.y -= offset.y()
                self.__class__.qt.QtGui.QCursor.setPos(center_global_posF.toPoint())
                cursor_global_posF = center_global_posF
        
        if cursor_global_pos != last_global_pos:
            try:
                screen_pos = self.mapFromGlobal(cursor_global_posF)
            except:
                screen_pos = self.mapFromGlobal(self.__class__.qt.QtCore.QPoint(int(cursor_global_posF.x()), int(cursor_global_posF.y())))
            
            screen_pos = glm.vec2(screen_pos.x(), screen_pos.y())
            global_pos = glm.vec2(cursor_global_posF.x(), cursor_global_posF.y())
            should_update = self._mouseMoveEvent(screen_pos, global_pos) or should_update

        if should_update:
            self.update()

def capture(self, save_path:str=None, viewport:tuple=None)->np.ndarray:
    self.makeCurrent()
    with self._before_PPE_fbo:
        scene = self.camera.scene
        clear_color = scene.fog.apply(scene.background.color, glm.vec3(0), glm.vec3(0,self.camera.far,0))
        with GLConfig.LocalConfig(clear_color=clear_color):
            with self.render_hints:
                self.renderer.render()

    image = None
    resolved = self._before_PPE_fbo.resolved
    if self._post_process_effects.has_valid:
        self._before_PPE_image = resolved.color_attachment(0)
        self._post_process_effects.screen_update_time = time.time()
        self._assign_values_to_PPEs()
        result_sampler = self._post_process_effects.apply(self._before_PPE_image)
        image = result_sampler.fbo.data(0)
    else:
        image = resolved.data(0)

    if viewport is None:
        viewport = GLConfig.buffered_viewport
    image = convert_to_image(image, viewport)
    
    if save_path is not None:
        folder_name = os.path.dirname(os.path.abspath(save_path))
        if not os.path.isdir(folder_name):
            os.makedirs(folder_name)

        cv2.imwrite(save_path, image)

    return image

def capture_video(self, save_path:str, viewport:tuple=None, fps:(float,int)=None)->VideoRecorder:
    ext_name = extname(save_path)
    if ext_name not in ["mp4", "avi"]:
        raise ValueError(f"not supported video type: .{ext_name}, only support .mp4 and .avi")
    
    fourcc = None
    if ext_name == "mp4":
        fourcc = cv2.VideoWriter_fourcc(*"mp4v")
    elif ext_name == "avi":
        fourcc = cv2.VideoWriter_fourcc(*"XVID")

    folder_path = os.path.dirname(os.path.abspath(save_path))
    if not os.path.isdir(folder_path):
        os.makedirs(folder_path)

    if fps is None:
        fps = self.smooth_fps

    if viewport is None:
        self.__class__.__base__.makeCurrent(self)
        try:
            viewport = GLConfig.buffered_viewport
        except:
            viewport = None

    video_recorder = VideoRecorder(self._video_recorders, save_path, fourcc, viewport, fps)
    self._video_recorders.append(video_recorder)

    return video_recorder

def closeEvent(self, close_event)->None:
    while self._video_recorders:
        video_recorder = self._video_recorders.pop()
        video_recorder.stop()

@property
def bloom(self):
    return self._post_process_effects["bloom"]

@bloom.setter
def bloom(self, flag:bool):
    self._post_process_effects["bloom"].enabled = flag

@property
def SSAO(self):
    return self._post_process_effects["SSAO"]

@SSAO.setter
def SSAO(self, flag:bool):
    self._post_process_effects["SSAO"].enabled = flag

@property
def tone_mapper(self):
    return self._post_process_effects["tone_mapper"]

@tone_mapper.setter
def tone_mapper(self, flag:bool):
    self._post_process_effects["tone_mapper"].enabled = flag

@property
def DOF(self):
    return self._post_process_effects["DOF"]

@DOF.setter
def DOF(self, flag:bool):
    self._post_process_effects["DOF"].enabled = flag

@property
def explosure_adaptor(self):
    return self._post_process_effects["explosure_adaptor"]

@explosure_adaptor.setter
def explosure_adaptor(self, flag:bool):
    self._post_process_effects["explosure_adaptor"].enabled = flag

@property
def FXAA(self):
    return self._post_process_effects["FXAA"]

@FXAA.setter
def FXAA(self, flag:bool):
    self._post_process_effects["FXAA"].enabled = flag

def init_QtScreen(cls):
    cls.qt = sys.modules[cls.__base__.__module__.split(".")[0]]
    cls._app = None
    cls._has_exec = False

    cls.__new__ = __new__
    cls.__init__ = __init__
    cls.update = update
    cls._assign_values_to_PPEs = _assign_values_to_PPEs
    cls._update = _update
    cls.sizeHint = sizeHint
    cls.camera = camera
    cls.samples = samples
    cls._set_samples = _set_samples
    cls.renderer = renderer
    cls.manipulator = manipulator
    cls.render_hints = render_hints
    cls.initializeGL = initializeGL
    cls.makeCurrent = makeCurrent
    cls.resizeGL = resizeGL
    cls._draw_to_before_PPE = _draw_to_before_PPE
    cls._mark_draw_calls = _mark_draw_calls
    cls.paintGL = paintGL
    cls._before_PPE_fbo = _before_PPE_fbo
    cls._mouse_parameters = _mouse_parameters
    cls._wheel_parameters = _wheel_parameters
    cls.mousePressEvent = mousePressEvent
    cls.mouseReleaseEvent = mouseReleaseEvent
    cls.mouseDoubleClickEvent = mouseDoubleClickEvent
    cls._mouseMoveEvent = _mouseMoveEvent
    cls.keyPressEvent = keyPressEvent
    cls.keyReleaseEvent = keyReleaseEvent
    cls._keyRepeateEvent = _keyRepeateEvent
    cls.wheelEvent = wheelEvent
    cls._calc_fps = _calc_fps
    cls.hide_cursor = hide_cursor
    cls.show_cursor = show_cursor
    cls.show = show
    cls.post_process_effects = post_process_effects
    cls.is_cursor_hiden = is_cursor_hiden
    cls.fps = fps
    cls.smooth_fps = smooth_fps
    cls.draw_calls = draw_calls
    cls.draw_points = draw_points
    cls.draw_lines = draw_lines
    cls.draw_meshes = draw_meshes
    cls.draw_patches = draw_patches
    cls.timerEvent = timerEvent
    cls.capture = capture
    cls.capture_video = capture_video
    cls.closeEvent = closeEvent
    cls.bloom = bloom
    cls.SSAO = SSAO
    cls.tone_mapper = tone_mapper
    cls.DOF = DOF
    cls.explosure_adaptor = explosure_adaptor
    cls.FXAA = FXAA

    return cls


from ..SceneNode import SceneNode
from ..callback_vec import callback_vec3

from glass.utils import checktype, di
from glass import GLConfig

import glm

class Light(SceneNode):

    def __init__(self, name:str=""):
        SceneNode.__init__(self, name)
        self._color:callback_vec3 = callback_vec3(1, 1, 1, callback=self._update_color)
        self._intensity:float = 1.0
        self._generate_shadows:bool = True
        self._rim_power:float = 0.3
        self._flats:set = set()

    def _set_transform_dirty(self, scenes):
        self._transform_dirty.update(scenes)
        return True

    @property
    def color(self):
        return self._color

    @color.setter
    @checktype
    def color(self, color:glm.vec3):
        if self._color == color:
            return
        
        self._color.r = color.r
        self._color.g = color.g
        self._color.b = color.b

    @property
    def intensity(self):
        return self._intensity

    @intensity.setter
    @checktype
    def intensity(self, intensity:float):
        if self._intensity == intensity:
            return

        self._intensity = intensity
        for flat in self._flats:
            flat.intensity = intensity

        self._update_scene_lights()

    @property
    def generate_shadows(self):
        return self._generate_shadows
    
    @generate_shadows.setter
    @checktype
    def generate_shadows(self, flag:bool):
        self._generate_shadows = flag
        self._update_generate_shadows()

    @property
    def rim_power(self):
        return self._rim_power
    
    @rim_power.setter
    @checktype
    def rim_power(self, rim_power:float):
        if self._rim_power == rim_power:
            return

        self._rim_power = rim_power
        for flat in self._flats:
            flat.rim_power = rim_power

        self._update_scene_lights()

    def _update_scene_lights(self):
        for scene in self.scenes:
            if self.__class__.__name__ == "PointLight":
                scene._point_lights.dirty = True
            elif self.__class__.__name__ == "DirLight":
                scene._dir_lights.dirty = True
            elif self.__class__.__name__ == "SpotLight":
                scene._spot_lights.dirty = True

    def _update_color(self):
        for flat in self._flats:
            flat.color = self._color

        self._update_scene_lights()

    def _update_generate_shadows(self):
        for flat in self._flats:
            flat.generate_shadows = self._generate_shadows

        self._update_scene_lights()

class FlatLight:

    def __init__(self, light:Light):
        self.depth_fbo_map = {}
        self.depth_map_handle = 0
        self._source_id = 0
        self.update(light)

    @property
    def depth_fbo(self):
        return self.depth_fbo_map.get(GLConfig.buffered_current_context, None)
    
    @depth_fbo.setter
    def depth_fbo(self, fbo):
        self.depth_fbo_map[GLConfig.buffered_current_context] = fbo

    def update(self, light:Light):
        self.color = light._intensity * light._color.flat
        self.generate_shadows = light._generate_shadows
        self.rim_power = light._rim_power
        
        self._source_id = id(light)
        light._flats.add(self)

    def before_del(self):
        if self._source_id == 0:
            return
        
        light = di(self._source_id)
        light._flats.remove(self)


from .utils import capacity_of
from .CSRMat import CSRMat

import threading

class Increment(CSRMat):
    def __init__(self, _list):
        len_list = len(_list)
        CSRMat.__init__(self, len_list, len_list+1, eye=True)
        self._is_changed = False
        self._lock = threading.Lock()

    def reset(self, _list):
        len_list = len(_list)
        self.indptr = list(range(len_list+1))
        self.data = _list
        self.col_indices = [self._cols-1] * len_list
        self.row_indices = list(range(len_list))

    def insert(self, index, value):
        with self._lock:
            if index < 0 or index > self.rows:
                raise IndexError(index)

            if isinstance(value, list):
                for i in range(len(value)):
                    self.insert_row(index+i)
                    self[index+i, self.cols-1] = value[i]
            else:
                self.insert_row(index)
                self[index, self.cols-1] = value

            self._is_changed = True

    def append(self, value):
        with self._lock:
            len_data = len(self.data)
            if isinstance(value, list):
                for sub_value in value:
                    self.data.append(sub_value)
                    len_data += 1

                    self.col_indices.append(self.cols-1)
                    self.row_indices.append(self.rows)
                    self.indptr.append(len_data)
            else:
                self.data.append(value)
                self.col_indices.append(self.cols-1)
                self.row_indices.append(self.rows)
                self.indptr.append(len_data)

            self._is_changed = True

    def update(self, index, value):
        if isinstance(index, list):
            len_index = len(index)
            if not isinstance(value, list):
                value = [value]*len_index
            for i in range(len_index):
                self.update(index[i], value[i])
            return

        with self._lock:
            if index < 0 or index >= self.rows:
                raise IndexError(index)

            start = self.indptr[index]
            stop = self.indptr[index+1]
            if stop - start == 1:
                self.data[start] = value
                self.col_indices[start] = self.cols-1
            else:
                self.reset_row(index)
                self[index, self.cols-1] = value

            self._is_changed = True

    def delete(self, start, stop=None):
        with self._lock:
            self.remove_row(start, stop)
            self._is_changed = True

    def clear(self):
        with self._lock:
            if self.rows > 0:
                self.indptr = [0]
                self.col_indices = []
                self.row_indices = []
                self.data = []
                self._is_changed = True

    def patch(self):
        with self._lock:
            self_cols = self.cols
            self_rows = self.rows
            
            result = \
            {
                "move": [],
                "update": [],
                "new_data": [],
                "old_size": self_cols-1,
                "new_size": self_rows,
                "old_capacity": capacity_of(self_cols-1),
                "new_capacity": capacity_of(self_rows)
            }

            if not self._is_changed:
                return result
            
            col_dict = {}
            for i, j, d in zip(self.row_indices, self.col_indices, self.data):                
                if j not in col_dict:
                    col_dict[j] = {}
                col_dict[j][i] = d

            self_cols_1 = self_cols-1

            last_i = -2
            update = None
            len_new_data = len(result["new_data"])
            for i, d in col_dict[self_cols_1].items():
                result["new_data"].append(d)
                len_new_data += 1
                if i != last_i+1:
                    update = \
                    {
                        "dest_start": i,
                        "src_start": len_new_data - 1,
                        "size": 1
                    }
                    result["update"].append(update)
                else:
                    update["size"] += 1

                last_i = i

            delta = 0
            
            move = None
            for j in sorted(col_dict.keys()):
                if len(col_dict[j]) != 1 or j == self_cols_1:
                    continue

                for i in col_dict[j]:
                    assert col_dict[j][i] == 1
                    current_delta = i - j
                    if current_delta == 0:
                        continue

                    if current_delta != delta:
                        move = \
                        {
                            "old_start": j,
                            "size": 1,
                            "new_start": i
                        }
                        result["move"].insert(0, move)
                        delta = current_delta
                    else:
                        move["size"] += 1

            CSRMat.__init__(self, self_rows, self_rows+1, eye=True)
            self._is_changed = False

            return result

    @property
    def is_changed(self):
        return self._is_changed


import math
import numpy as np
from abc import ABC, abstractmethod

class WeightedIntegrator(ABC):

    def __init__(self):
        self.__ortho_polys = {0: np.array([1])}
        self.__nodes = {0: np.array([])}
        self.__weights = {0: np.array([])}

    @abstractmethod
    def integral_base_with_weight(self, n):
        pass

    def integral_func_with_weight(self, func, n):
        x = self.nodes(n)
        w = self.weights(n)
        result = 0
        for i in range(n):
            result += w[i] * func(x[i])

        return result

    def nodes(self, n):
        if n in self.__nodes:
            return self.__nodes[n]
        
        self.__nodes[n] = np.sort(np.real(np.roots(self._ortho_poly(n))))
        return self.__nodes[n]
    
    def weights(self, n):
        if n in self.__weights:
            return self.__weights[n]

        I = []
        for i in range(n):
            I.append(self.integral_base_with_weight(i))
        I = np.array(I)

        x = self.nodes(n)
        V = np.vander(x, increasing=True).T
        invV = np.linalg.inv(V)
        self.__weights[n] = np.dot(invV, I)
        return self.__weights[n]

    def _ortho_poly(self, n):
        if n in self.__ortho_polys:
            return self.__ortho_polys[n]
        
        xn = np.array([1] + [0]*n)
        poly = xn
        for i in range(n):
            pi = self._ortho_poly(i)
            item_num = np.polymul(xn, pi)
            num = 0
            for j in range(len(item_num)):
                num += item_num[j]*self.integral_base_with_weight(len(item_num)-j-1)

            item_dem = np.polymul(pi, pi)
            dem = 0
            for j in range(len(item_dem)):
                dem += item_dem[j]*self.integral_base_with_weight(len(item_dem)-j-1)

            poly = np.polysub(poly, num/dem*pi)

        self.__ortho_polys[n] = poly
        return poly

class cosxIntegrator(WeightedIntegrator):

    def __init__(self):
        WeightedIntegrator.__init__(self)
        self.__I2n_values = {0: 2}

    def integral_base_with_weight(self, n):
        if n & 1:
            return 0
        
        half_n = n//2
        if half_n in self.__I2n_values:
            return self.__I2n_values[half_n]
        
        self.__I2n_values[half_n] = 2*(math.pi/2)**n - n*(n-1)*self.integral_base_with_weight(n - 2)
        return self.__I2n_values[half_n]



import numpy as np
import queue
import threading
import math
import cv2
import time

from glass.utils import di

def convert_to_image(data:np.ndarray, viewport:tuple)->np.ndarray:
    if viewport[0] != 0 or \
       viewport[1] != 0 or \
       viewport[2] != data.shape[1] or \
       viewport[3] != data.shape[0]:
        width_pad = max(viewport[0] + viewport[2] - data.shape[1], 0)
        height_pad = max(viewport[1] + viewport[3] - data.shape[0], 0)
        if len(data.shape) == 2:
            if width_pad > 0 or height_pad > 0:
                data = np.pad(data, ((height_pad,0), (0,width_pad)))
            data = data[viewport[1]:viewport[1]+viewport[3], viewport[0]:viewport[0]+viewport[2]]
        else:
            if width_pad > 0 or height_pad > 0:
                data = np.pad(data, ((height_pad,0), (0,width_pad), (0, 0)))
            data = data[viewport[1]:viewport[1]+viewport[3], viewport[0]:viewport[0]+viewport[2], :]

    image = data
    if "float" in str(data.dtype):
        image = np.clip(255*data, 0, 255)

    image = image.astype(np.uint8)
    image = cv2.flip(image, 0)

    channels = (1 if len(image.shape) < 3 else image.shape[2])
    if channels == 3:
        image = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)
    elif channels == 4:
        image = cv2.cvtColor(image, cv2.COLOR_RGBA2BGR)

    return image

class VideoRecorder:
    def __init__(self, screen, file_name:str, fourcc:list, viewport:tuple, fps:(float,int))->None:
        self._screen_id = id(screen)

        self._file_name = file_name
        self._fourcc = fourcc

        self._last_frame = None
        self._frame_queue = queue.Queue()
        self._stop_time = float("inf")
        self._writing_thread = threading.Thread(target=self._writing_loop)
        self._started = False
        if viewport is not None and fps is not None:
            self._start(viewport, fps)

    def _start(self, viewport, fps):
        if self._started:
            return
        
        if fps == 0:
            fps = 24

        self._viewport = viewport
        self._dt = 1/fps
        self._cv_video_writer = cv2.VideoWriter(self._file_name, self._fourcc, fps, (viewport[2], viewport[3]))
        self._start_time = time.time()
        self._writing_thread.start()
        self._started = True

    def __del__(self)->None:
        self.stop()

    def stop(self)->None:
        if math.isinf(self._stop_time):
            self._stop_time = time.time()
            self._frame_queue.put(None)

    def _writing_loop(self)->None:
        t = self._start_time
        while t < self._stop_time:
            frame = self._frame_queue.get()
            if frame is None:
                if self._last_frame is None:
                    break

                while t < self._stop_time:
                    self._cv_video_writer.write(self._last_frame[1])
                    t += self._dt

                break
            image = convert_to_image(frame[1], self._viewport)

            if self._last_frame is None:
                self._last_frame = (frame[0], image)

            while t <= min(self._stop_time, frame[0]):
                self._cv_video_writer.write(self._last_frame[1])
                t += self._dt

            self._last_frame = (frame[0], image)

        self._cv_video_writer.release()
        try:
            screen = di(self._screen_id)
            screen._screen_video_writers.remove(self)
        except:
            pass


import os

import cv2
import numpy as np
import glm
from OpenGL import GL
import OpenGL.GL.ARB.bindless_texture as bt
import time
from datetime import datetime

from .FBOAttachment import FBOAttachment
from .GLInfo import GLInfo
from .utils import checktype, cat, modify_time, md5s, relative_path, is_text_file
from .helper import get_external_format, width_adapt, get_dtype, get_channels
from .ImageLoader import ImageLoader
from .Indices import Indices
from .Vertices import Vertices, Vertex
from .GLConfig import GLConfig
from .GlassConfig import GlassConfig

class sampler2D(FBOAttachment):

    _default_internal_format = GL.GL_RGBA32F
    _default_filter_min = GL.GL_LINEAR
    _default_filter_mag = GL.GL_LINEAR
    _default_filter_mipmap = GL.GL_LINEAR

    _sampler2D_map = {}
    _should_update = False
    __shadertoy_template_content = ""
    __shadertoy_template_filename = os.path.dirname(os.path.abspath(__file__)) + "/glsl/shadertoy_template.glsl"

    _basic_info = \
    {
        "gen_func": GL.glGenTextures,
        "bind_func": GL.glBindTexture,
        "del_func": GL.glDeleteTextures,
        "target_type": GL.GL_TEXTURE_2D,
        "binding_type": GL.GL_TEXTURE_BINDING_2D,
        "need_number": True,
    }

    @checktype
    def __init__(self, image:(str,np.ndarray)=None, width:int=None, height:int=None, internal_format:GLInfo.internal_formats=None):
        FBOAttachment.__init__(self)
        
        self._handle = 0
        self._image = None
        self._width = 0
        self._height = 0
        self._internal_format = self.__class__._default_internal_format

        self._wrap_s = GL.GL_REPEAT
        self._wrap_t = GL.GL_REPEAT
        self._filter_min = self.__class__._default_filter_min
        self._filter_mag = self.__class__._default_filter_mag
        self._filter_mipmap = self.__class__._default_filter_mipmap
        self._border_color = glm.vec4(0, 0, 0, 1)

        self._image_changed = True
        self._wrap_s_changed = True
        self._wrap_t_changed = True
        self._filter_min_changed = True
        self._filter_mag_changed = True
        self._border_color_changed = False

        # shadertoy
        self._shadertoy_program = None
        self._shadertoy_path = None
        self._shadertoy_start_time = 0
        self._shadertoy_last_frame_time = 0
        self._shadertoy_frame_index = 0
        self._should_update_shadertoy = True
        self.__frame_vertices = None
        self.__frame_indices = None

        if image is not None:
            self.image = image

        if width is not None:
            self.width = width

        if height is not None:
            self.height = height

        if internal_format is not None:
            self.internal_format = internal_format

    def __hash__(self):
        return id(self)

    def __deepcopy__(self, memo):
        result = self.__class__()
        result._image = self._image
        result._width = self._width
        result._height = self._height
        result._internal_format = self._internal_format

        result._wrap_s = self._wrap_s
        result._wrap_t = self._wrap_t
        result._filter_min = self._filter_min
        result._filter_mag = self._filter_mag
        result._filter_mipmap = self._filter_mipmap
        result._border_color = self._border_color
        result._dynamic = self._dynamic

        result._shadertoy_program = None
        result._shadertoy_path = self._shadertoy_path
        result._shadertoy_start_time = self._shadertoy_start_time
        result._shadertoy_last_frame_time = self._shadertoy_last_frame_time
        result._shadertoy_frame_index = self._shadertoy_frame_index
        if self._shadertoy_program is not None:
            result.__init_shadertoy(self._shadertoy_path)

        return result

    def __del__(self):
        if self._handle != 0:
            self._handle = 0

        FBOAttachment.__del__(self)

    @property
    def is_shadertoy(self)->bool:
        return bool(self._shadertoy_path)

    @classmethod
    def load(cls, file_name:str, internal_format:GLInfo.internal_formats=None):
        if not os.path.isfile(file_name):
            raise FileNotFoundError("not a valid image file: " + file_name)
        
        file_name = os.path.abspath(file_name).replace("\\", "/")
        if file_name not in sampler2D._sampler2D_map:
            sampler = cls(file_name, internal_format=internal_format)
            sampler._dynamic = False
            sampler2D._sampler2D_map[file_name] = sampler

        return sampler2D._sampler2D_map[file_name]

    @property
    def handle(self):
        if not bt.glGetTextureHandleARB:
            return 0
        
        self.bind()
        if self._handle == 0:
            self._handle = bt.glGetTextureHandleARB(self._id)
            if self._handle == 0:
                raise RuntimeError(f"failed to create sampler2D {self._id}'s handle")
            bt.glMakeTextureHandleResidentARB(self._handle)            
            self._dynamic = False

        return self._handle

    def bind(self, update_fbo:bool=False, force_update_image:bool=False):
        if self._should_update_shadertoy and self._shadertoy_program is not None and (self.fbo.context == 0 or self.fbo.context == GLConfig.buffered_current_context):
            self.__update_shadertoy()

        FBOAttachment.bind(self, update_fbo)

        if self._wrap_s_changed:
            GL.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_S, self._wrap_s)
            self._wrap_s_changed = False
        if self._wrap_t_changed:
            GL.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_T, self._wrap_t)
            self._wrap_t_changed = False
        if self._border_color_changed:
            GL.glTexParameterfv(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_BORDER_COLOR, glm.value_ptr(self._border_color))
            self._border_color_changed = False

        if self._filter_min_changed:
            GL.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MIN_FILTER, self._get_filter_min())
            self._filter_min_changed = False
        if self._filter_mag_changed:
            GL.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MAG_FILTER, self._filter_mag)
            self._filter_mag_changed = False

        generated_mipmap = False
        if force_update_image or self._image_changed:
            external_format = get_external_format(self._internal_format)
            width_adapt(self._width)
            GL.glTexImage2D(GL.GL_TEXTURE_2D, 0, self._internal_format,
                            self._width, self._height, 0, external_format, self.dtype, self._image)
            
            if not update_fbo and self._filter_mipmap is not None and self._image is not None:
                self.generate_mipmap()
                generated_mipmap = True

            self._image_changed = False

        if not update_fbo and not generated_mipmap and not self._fbo_image_generated_mipmap:
            if self._filter_mipmap is not None:
                self.generate_mipmap()

            self._fbo_image_generated_mipmap = True

    def generate_mipmap(self):
        FBOAttachment.bind(self)
        GL.glGenerateMipmap(GL.GL_TEXTURE_2D)

    def clear(self):
        if self._id == 0:
            return

        self.bind()
        external_format = get_external_format(self._internal_format)
        dtype = get_dtype(self._internal_format)
        GL.glClearTexImage(self._id, 0, external_format, dtype, None)

    @property
    def width(self):
        return self._width

    @width.setter
    @FBOAttachment.param_setter
    def width(self, width:int):
        self._width = width
        if self._image is not None:
            if len(self._image.shape) > 2:
                cv2.resize(self._image, (width, self._image.shape[0], self._image.shape[2]))
            else:
                cv2.resize(self._image, (width, self._image.shape[0]))

        self._image_changed = True

    @property
    def height(self):
        return self._height

    @height.setter
    @FBOAttachment.param_setter
    def height(self, height:int):
        self._height = height
        if self._image is not None:
            if len(self._image.shape) > 2:
                cv2.resize(self._image, (self._image.shape[1], height, self._image.shape[2]))
            else:
                cv2.resize(self._image, (self._image.shape[1], height))

        self._image_changed = True

    @property
    def dtype(self):
        return get_dtype(self._internal_format)

    @property
    def internal_format(self):
        return self._internal_format

    def _set_internal_format(self, internal_format):
        self._internal_format = internal_format

        if self._image is not None:
            old_channels = self._image.shape[2] if len(self._image.shape) > 2 else 1
            new_channels = get_channels(internal_format)
            convert_map = \
            {
                (1,3): cv2.COLOR_GRAY2RGB,
                (1,4): cv2.COLOR_GRAY2RGBA,
                (3,1): cv2.COLOR_RGB2GRAY,
                (3,4): cv2.COLOR_RGB2RGBA,
                (4,1): cv2.COLOR_RGBA2GRAY,
                (4,3): cv2.COLOR_RGBA2RGB
            }
            if (old_channels,new_channels) in convert_map:
                conversion = convert_map[(old_channels,new_channels)]
                self._image = cv2.cvtColor(self._image, conversion)
            elif old_channels == 1:
                if len(self._image.shape) == 2:
                    self._image = np.expand_dims(self._image, axis=2)
                self._image = self._image.repeat(new_channels, axis=2)
            elif new_channels > old_channels:
                self._image = np.append(self._image, np.zeros((self._image.shape[0], self._image.shape[1], new_channels-old_channels), dtype=self._image.dtype), axis=2)
            elif new_channels < old_channels:
                self._image = self._image[:, :, :new_channels]

        if self._image is not None:
            old_np_dtype = self._image.dtype
            new_np_dtype = GLInfo.dtype_map[get_dtype(self._internal_format)]
            if old_np_dtype != new_np_dtype:
                if "int" in str(old_np_dtype) and "float" in str(new_np_dtype):
                    self._image = (self._image / 255).astype(new_np_dtype)
                elif "float" in str(old_np_dtype) and "int" in str(new_np_dtype):
                    self._image = (self._image * 255).astype(new_np_dtype)
                else:
                    self._image = self._image.astype(new_np_dtype)

        self._image_changed = True

    @internal_format.setter
    @FBOAttachment.param_setter
    def internal_format(self, internal_format:GLInfo.internal_formats):
        self._set_internal_format(internal_format)

    @property
    def wrap_s(self):
        return self._wrap_s

    @wrap_s.setter
    @FBOAttachment.param_setter
    def wrap_s(self, wrap_type:GLInfo.wrap_types):
        self._wrap_s = wrap_type
        self._wrap_s_changed = True

    @property
    def wrap_t(self):
        return self._wrap_t

    @wrap_t.setter
    @FBOAttachment.param_setter
    def wrap_t(self, wrap_type:GLInfo.wrap_types):
        self._wrap_t = wrap_type
        self._wrap_t_changed = True

    @property
    def wrap(self):
        return self._wrap_s, self._wrap_t

    @wrap.setter
    @FBOAttachment.param_setter
    def wrap(self, wrap_type):
        if wrap_type in GLInfo.wrap_types:
            self.wrap_s = wrap_type
            self.wrap_t = wrap_type
        else:
            self.wrap_s = wrap_type[0]
            self.wrap_t = wrap_type[1]

    @property
    def filter_min(self):
        return self._filter_min

    @filter_min.setter
    @FBOAttachment.param_setter
    def filter_min(self, filter_type:GLInfo.filter_types):
        self._filter_min = filter_type
        self._filter_min_changed = True

    @property
    def filter_mag(self):
        return self._filter_mag

    @filter_mag.setter
    @FBOAttachment.param_setter
    def filter_mag(self, filter_type:GLInfo.filter_types):
        self._filter_mag = filter_type
        self._filter_mag_changed = True

    @property
    def filter_mipmap(self):
        return self._filter_mipmap

    @filter_mipmap.setter
    @FBOAttachment.param_setter
    def filter_mipmap(self, filter_type:GLInfo.filter_types):
        self._filter_mipmap = filter_type
        self._filter_min_changed = True
        self._filter_mag_changed = True

    def _get_filter_min(self):
        if self._filter_mipmap is None:
            return self._filter_min
        else:
            return GLInfo.mipmap_filter_map[(self._filter_min, self._filter_mipmap)]
        
    @property
    def filter(self):
        return self._filter_min, self._filter_mag, self._filter_mipmap

    @filter.setter
    @FBOAttachment.param_setter
    def filter(self, filter_type):
        if filter_type in GLInfo.filter_types:
            self.filter_min = filter_type
            self.filter_mag = filter_type
            self.filter_mipmap = filter_type
        else:
            self.filter_min = filter_type[0]
            self.filter_mag = filter_type[1]
            if len(filter_type) > 2:
                self.filter_mipmap = filter_type[2]

    @property
    def border_color(self):
        return self._border_color

    @border_color.setter
    @FBOAttachment.param_setter
    def border_color(self, color:(glm.vec3, glm.vec4)):
        if isinstance(color, glm.vec3):
            color = glm.vec4(color, 1)

        self._border_color = color
        self._border_color_changed = True

    @property
    def image(self):
        if self._fbo_image_changed:
            self._image = self._fbo.data(self._fbo_attach_point)
            self._image_changed = False
            self._fbo_image_changed = False

        return self._image

    @image.setter
    @FBOAttachment.param_setter
    def image(self, image:(np.ndarray,str)):
        is_shadertoy = False
        if isinstance(image, str):
            try:
                image = ImageLoader.load(image)
            except ValueError as e:
                if is_text_file(image):
                    is_shadertoy = True
                else:
                    raise e

        self._shadertoy_program = None
        self._shadertoy_path = None
        self._shadertoy_start_time = 0
        self._shadertoy_last_frame_time = 0
        self._shadertoy_frame_index = 0

        if not is_shadertoy:
            self._image = image
            self._width = self._image.shape[1]
            self._height = self._image.shape[0]

            channels = self._image.shape[2] if len(self._image.shape) > 2 else 1
            self._internal_format = GLInfo.internal_formats_map[self._image.dtype][channels]
            self._image_changed = True
        else:
            self.__init_shadertoy(image)

    @checktype
    def malloc(self, width:int, height:int, samples:int=None, layers:int=None, internal_format:GLInfo.internal_formats=None):        
        self.width = width
        self.height = height

        if internal_format is None:
            internal_format = self.internal_format

        if internal_format is None:
            internal_format = self.__class__._default_internal_format

        self.internal_format = internal_format

    @property
    def is_completed(self):
        return (self._width > 0 and self._height > 0)
    
    def __getitem__(self, name:str):
        if self._shadertoy_program is None:
            raise RuntimeError("not a shader defined sampler2D")
        
        return self._shadertoy_program[name]
    
    def __setitem__(self, name:str, value):
        if self._shadertoy_program is None:
            raise RuntimeError("not a shader defined sampler2D")
        
        self._shadertoy_program[name] = value

    @property
    def samples(self):
        return None
    
    @samples.setter
    def samples(self, samples:int):
        pass

    def __init_shadertoy(self, shader_path):
        from .ShaderProgram import ShaderProgram
        from .FBO import FBO

        if self._width == 0:
            self._width = 800
        if self._height == 0:
            self._height = 800

        self._shadertoy_path = shader_path
        if not os.path.isfile(shader_path):
            raise FileNotFoundError(shader_path)
        
        shader_path = os.path.abspath(shader_path).replace("\\", "/")
        base_name = os.path.basename(shader_path)
        out_file_name = GlassConfig.cache_folder + "/" + base_name + "_" + md5s(shader_path) + ".glsl"
        if modify_time(sampler2D.__shadertoy_template_filename) > modify_time(out_file_name):
            out_file = open(out_file_name, "w")
            out_file.write(sampler2D.__shadertoy_template(shader_path))
            out_file.close()

        program = ShaderProgram()
        program.compile(os.path.dirname(os.path.abspath(__file__)) + "/glsl/draw_frame.vs")
        program.compile(out_file_name, GL.GL_FRAGMENT_SHADER)

        fbo = FBO(self.width, self.height)
        fbo.attach(0, attachment=self)

        self._shadertoy_program = program

    @staticmethod
    def __shadertoy_template(file_name):
        if not sampler2D.__shadertoy_template_content:
            sampler2D.__shadertoy_template_content = cat(sampler2D.__shadertoy_template_filename)
        
        rel_path = relative_path(file_name, GlassConfig.cache_folder)
        return sampler2D.__shadertoy_template_content.replace("{file_name}", rel_path)

    @property
    def _frame_vertices(self):
        if self.__frame_vertices is None:
            self.__frame_vertices = Vertices()
            self.__frame_vertices[0] = Vertex(position=glm.vec2(-1, -1))
            self.__frame_vertices[1] = Vertex(position=glm.vec2(1, -1))
            self.__frame_vertices[2] = Vertex(position=glm.vec2(1, 1))
            self.__frame_vertices[3] = Vertex(position=glm.vec2(-1, 1))

        return self.__frame_vertices
    
    @property
    def _frame_indices(self):
        if self.__frame_indices is None:
            self.__frame_indices = Indices()
            self.__frame_indices[0] = glm.uvec3(0, 1, 2)
            self.__frame_indices[1] = glm.uvec3(0, 2, 3)

        return self.__frame_indices

    def __update_shadertoy(self):
        self._should_update_shadertoy = False

        is_dynamic = (self._shadertoy_program["iTime"].location != -1 or \
                      self._shadertoy_program["iTimeDelta"].location != -1 or \
                      self._shadertoy_program["iFrameRate"].location != -1 or \
                      self._shadertoy_program["iFrame"].location != -1 or \
                      self._shadertoy_program["iChannelTime"].location != -1 or \
                      self._shadertoy_program["iDate"].location != -1)
        if is_dynamic:
            sampler2D._should_update = True

        current_time = time.time()
        now = datetime.now()

        if self._shadertoy_start_time == 0:
            self._shadertoy_start_time = current_time
        if self._shadertoy_last_frame_time == 0:
            self._shadertoy_last_frame_time = current_time
        
        time_delta = current_time - self._shadertoy_last_frame_time
        fps = 60
        if time_delta > 0:
            fps = 1/time_delta
        t = current_time - self._shadertoy_start_time
        
        with GLConfig.LocalConfig(cull_face=None, polygon_mode=GL.GL_FILL):
            with self.fbo:
                resolution = glm.vec3(self._width, self._height, 1)
                self._shadertoy_program["iResolution"] = resolution
                self._shadertoy_program["iChannelResolution"] = [resolution] * 4
                self._shadertoy_program["iTime"] = t
                self._shadertoy_program["iTimeDelta"] = time_delta
                self._shadertoy_program["iFrameRate"] = fps
                self._shadertoy_program["iFrame"] = self._shadertoy_frame_index
                self._shadertoy_program["iChannelTime"] = [t, t, t, t]
                self._shadertoy_program["iDate"] = glm.vec4(now.year, now.month, now.day, now.second + now.microsecond/1000)
                self._shadertoy_program["iSampleRate"] = 44100
                
                self._shadertoy_last_frame_time = current_time
                self._shadertoy_frame_index += 1
                self._shadertoy_program.draw_triangles(self._frame_vertices, self._frame_indices)
        
        self._should_update_shadertoy = True



from abc import ABC, abstractmethod
from glass import sampler2D
from glass.utils import checktype

class PostProcessEffect(ABC):

    def __init__(self):
        self._should_update = False
        self._enabled = True
        self._screen_update_time = 0
        
        self.depth_map = None
        self.view_pos_map = None
        self.view_normal_map = None
        self.camera = None

    def __bool__(self)->bool:
        return self._enabled

    @abstractmethod
    def apply(self, screen_image:sampler2D)->sampler2D:
        pass

    @abstractmethod
    def draw_to_active(self, screen_image:sampler2D)->None:
        pass

    @abstractmethod
    def need_pos_info(self)->bool:
        return False

    @property
    def enabled(self)->bool:
        return self._enabled
    
    @enabled.setter
    @checktype
    def enabled(self, flag:bool):
        self._enabled = flag

    @property
    def should_update(self)->bool:
        return (self._should_update and self._enabled)
    
    @should_update.setter
    @checktype
    def should_update(self, flag:bool):
        self._should_update = flag

    @property
    def screen_update_time(self):
        return self._screen_update_time
    
    @screen_update_time.setter
    @checktype
    def screen_update_time(self, screen_update_time:float):
        self._screen_update_time = screen_update_time


from .PostProcessEffect import PostProcessEffect
from ..Frame import Frame

from glass import FBO, ShaderProgram, sampler2D, GLConfig
from glass.utils import checktype
from glass.ShaderStorageBlock import ShaderStorageBlock

from OpenGL import GL
import time
import os

class DOFEffect(PostProcessEffect):

    class CurrentFocus(ShaderStorageBlock.HostClass):
        def __init__(self):
            ShaderStorageBlock.HostClass.__init__(self)

            self._current_focus = 0

        @property
        def current_focus(self):
            return self._current_focus
        
        @current_focus.setter
        @ShaderStorageBlock.HostClass.not_const
        def current_focus(self, focus:float):
            self._current_focus = focus

    def __init__(self):
        PostProcessEffect.__init__(self)
        self.current_focus = DOFEffect.CurrentFocus()

        self._horizontal_fbo = None
        self._vertical_fbo = None
        self._program = None

    def need_pos_info(self) -> bool:
        return True

    @property
    def horizontal_fbo(self):
        if self._horizontal_fbo is None:
            self._horizontal_fbo = FBO()
            self._horizontal_fbo.attach(0, sampler2D, GL.GL_RGBA32F)
        return self._horizontal_fbo

    @property
    def vertical_fbo(self):
        if self._vertical_fbo is None:
            self._vertical_fbo = FBO()
            self._vertical_fbo.attach(0, sampler2D, GL.GL_RGBA32F)
        return self._vertical_fbo

    @property
    def program(self):
        if self._program is None:
            self._program = ShaderProgram()
            self._program.compile(Frame.draw_frame_vs)
            self._program.compile(os.path.dirname(os.path.abspath(__file__)) + "/../glsl/PostProcessEffects/dof.fs")
            self._program["CurrentFocus"].bind(self.current_focus)
        return self._program
    
    def apply(self, screen_image:sampler2D)->sampler2D:
        self.horizontal_fbo.resize(screen_image.width, screen_image.height)
        self.vertical_fbo.resize(screen_image.width, screen_image.height)

        with GLConfig.LocalConfig(cull_face=None, polygon_mode=GL.GL_FILL):
            self.program["camera"] = self.camera
            self.program["view_pos_map"] = self.view_pos_map
            self.program["fps"] = self.camera.screen.smooth_fps
            with self.horizontal_fbo:
                self.program["screen_image"] = screen_image
                self.program["horizontal"] = True
                self.program.draw_triangles(Frame.vertices, Frame.indices)

            with self.vertical_fbo:
                self.program["screen_image"] = self.horizontal_fbo.color_attachment(0)
                self.program["horizontal"] = False
                self.program.draw_triangles(Frame.vertices, Frame.indices)

        return self.vertical_fbo.color_attachment(0)

    def draw_to_active(self, screen_image:sampler2D)->None:
        self.horizontal_fbo.resize(screen_image.width, screen_image.height)
        self.vertical_fbo.resize(screen_image.width, screen_image.height)

        with GLConfig.LocalConfig(cull_face=None, polygon_mode=GL.GL_FILL):
            self.program["camera"] = self.camera
            self.program["view_pos_map"] = self.view_pos_map
            self.program["fps"] = self.camera.screen.smooth_fps
            with self.horizontal_fbo:
                self.program["screen_image"] = screen_image
                self.program["horizontal"] = True
                self.program.draw_triangles(Frame.vertices, Frame.indices)

            GLConfig.clear_buffers()
            self.program["screen_image"] = self.horizontal_fbo.color_attachment(0)
            self.program["horizontal"] = False
            self.program.draw_triangles(Frame.vertices, Frame.indices)
    
    @property
    def should_update(self) -> bool:
        if not self._enabled:
            return False
        
        if not self.camera.lens.auto_focus:
            return False

        return (self.screen_update_time == 0 or time.time()-self.screen_update_time <= self.camera.lens.focus_change_time+1)
    
    @should_update.setter
    @checktype
    def should_update(self, flag:bool):
        pass
